#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Restricted Open-shell Hartree-Fock
'''

from functools import reduce
import numpy
import pyscf.gto
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import hf
from pyscf.scf import uhf
from pyscf import __config__

WITH_META_LOWDIN = getattr(__config__, 'scf_analyze_with_meta_lowdin', True)
MO_BASE = getattr(__config__, 'MO_BASE', 1)


def init_guess_by_minao(mol):
    dm = hf.init_guess_by_minao(mol)
    dm = numpy.array((dm*.5, dm*.5))
    if hasattr(dm, 'mo_coeff'):
        dm = lib.tag_array(dm, mo_coeff=dm.mo_coeff, mo_occ=dm.mo_occ)
    return dm

def init_guess_by_atom(mol):
    dm = hf.init_guess_by_atom(mol)
    dm = numpy.array((dm*.5, dm*.5))
    if hasattr(dm, 'mo_coeff'):
        dm = lib.tag_array(dm, mo_coeff=dm.mo_coeff, mo_occ=dm.mo_occ)
    return dm

def init_guess_by_sap(mol, sap_basis, **kwargs):
    dm = hf.init_guess_by_sap(mol, sap_basis, **kwargs)
    dm = numpy.array((dm*.5, dm*.5))
    if hasattr(dm, 'mo_coeff'):
        dm = lib.tag_array(dm, mo_coeff=dm.mo_coeff, mo_occ=dm.mo_occ)
    return dm

init_guess_by_huckel = uhf.init_guess_by_huckel
init_guess_by_mod_huckel = uhf.init_guess_by_mod_huckel

def init_guess_by_chkfile(mol, chkfile_name, project=None):
    '''Read SCF chkfile and make the density matrix for ROHF initial guess.

    Kwargs:
        project : None or bool
            Whether to project chkfile's orbitals to the new basis.  Note when
            the geometry of the chkfile and the given molecule are very
            different, this projection can produce very poor initial guess.
            In PES scanning, it is recommended to switch off project.

            If project is set to None, the projection is only applied when the
            basis sets of the chkfile's molecule are different to the basis
            sets of the given molecule (regardless whether the geometry of
            the two molecules are different).  Note the basis sets are
            considered to be different if the two molecules are derived from
            the same molecule with different ordering of atoms.
    '''
    dm = uhf.init_guess_by_chkfile(mol, chkfile_name, project)
    mo_coeff = dm.mo_coeff[0]
    mo_occ = dm.mo_occ[0] + dm.mo_occ[1]
    return lib.tag_array(dm, mo_coeff=mo_coeff, mo_occ=mo_occ)

def get_fock(mf, h1e=None, s1e=None, vhf=None, dm=None, cycle=-1, diis=None,
             diis_start_cycle=None, level_shift_factor=None, damp_factor=None,
             fock_last=None):
    '''Build fock matrix based on Roothaan's effective fock.
    See also :func:`get_roothaan_fock`
    '''
    if h1e is None: h1e = mf.get_hcore()
    if s1e is None: s1e = mf.get_ovlp()
    if vhf is None: vhf = mf.get_veff(mf.mol, dm)
    if dm is None: dm = mf.make_rdm1()
    if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
        dm = numpy.array((dm*.5, dm*.5))
# To Get orbital energy in get_occ, we saved alpha and beta fock, because
# Roothaan effective Fock cannot provide correct orbital energy with `eig`
# TODO, check other treatment  J. Chem. Phys. 133, 141102
    focka = h1e + vhf[0]
    fockb = h1e + vhf[1]
    f = get_roothaan_fock((focka,fockb), dm, s1e)
    if cycle < 0 and diis is None:  # Not inside the SCF iteration
        return f

    if diis_start_cycle is None:
        diis_start_cycle = mf.diis_start_cycle
    if level_shift_factor is None:
        level_shift_factor = mf.level_shift
    if damp_factor is None:
        damp_factor = mf.damp

    dm_tot = dm[0] + dm[1]
    if 0 <= cycle < diis_start_cycle-1 and abs(damp_factor) > 1e-4 and fock_last is not None:
        raise NotImplementedError('ROHF Fock-damping')
    if diis and cycle >= diis_start_cycle:
        f = diis.update(s1e, dm_tot, f, mf, h1e, vhf, f_prev=fock_last)
    if abs(level_shift_factor) > 1e-4:
        f = hf.level_shift(s1e, dm_tot*.5, f, level_shift_factor)
    f = lib.tag_array(f, focka=focka, fockb=fockb)
    return f

def get_roothaan_fock(focka_fockb, dma_dmb, s):
    '''Roothaan's effective fock.
    Ref. http://www-theor.ch.cam.ac.uk/people/ross/thesis/node15.html

    ======== ======== ====== =========
    space     closed   open   virtual
    ======== ======== ====== =========
    closed      Fc      Fb     Fc
    open        Fb      Fc     Fa
    virtual     Fc      Fa     Fc
    ======== ======== ====== =========

    where Fc = (Fa + Fb) / 2

    Returns:
        Roothaan effective Fock matrix
    '''
    nao = s.shape[0]
    focka, fockb = focka_fockb
    dma, dmb = dma_dmb
    fc = (focka + fockb) * .5
# Projector for core, open-shell, and virtual
    pc = numpy.dot(dmb, s)
    po = numpy.dot(dma-dmb, s)
    pv = numpy.eye(nao) - numpy.dot(dma, s)
    fock  = reduce(numpy.dot, (pc.conj().T, fc, pc)) * .5
    fock += reduce(numpy.dot, (po.conj().T, fc, po)) * .5
    fock += reduce(numpy.dot, (pv.conj().T, fc, pv)) * .5
    fock += reduce(numpy.dot, (po.conj().T, fockb, pc))
    fock += reduce(numpy.dot, (po.conj().T, focka, pv))
    fock += reduce(numpy.dot, (pv.conj().T, fc, pc))
    fock = fock + fock.conj().T
    fock = lib.tag_array(fock, focka=focka, fockb=fockb)
    return fock


def get_occ(mf, mo_energy=None, mo_coeff=None):
    '''Label the occupancies for each orbital.

    Examples:

    >>> mol = gto.M(atom='H 0 0 0; O 0 0 1.1', spin=1)
    >>> mf = scf.hf.SCF(mol)
    >>> energy = numpy.array([-10., -1., 1, -2., 0, -3])
    >>> mf.get_occ(energy)
    array([2, 2, 0, 2, 1, 2])
    '''

    if mo_energy is None: mo_energy = mf.mo_energy
    if getattr(mo_energy, 'mo_ea', None) is not None:
        mo_ea = mo_energy.mo_ea
        mo_eb = mo_energy.mo_eb
    else:
        mo_ea = mo_eb = mo_energy
    nmo = mo_ea.size
    if getattr(mf, 'nelec', None) is None:
        nelec = mf.mol.nelec
    else:
        nelec = mf.nelec
    if nelec[0] > nelec[1]:
        nocc, ncore = nelec
    else:
        ncore, nocc = nelec
    nopen = nocc - ncore
    mo_occ = _fill_rohf_occ(mo_energy, mo_ea, mo_eb, ncore, nopen)

    if mf.verbose >= logger.INFO and nocc < nmo and ncore > 0:
        ehomo = max(mo_energy[mo_occ> 0])
        elumo = min(mo_energy[mo_occ==0])
        if ehomo+1e-3 > elumo:
            logger.warn(mf, 'HOMO %.15g >= LUMO %.15g', ehomo, elumo)
        else:
            logger.info(mf, '  HOMO = %.15g  LUMO = %.15g', ehomo, elumo)
        if nopen > 0 and mf.verbose >= logger.DEBUG:
            core_idx = mo_occ == 2
            open_idx = mo_occ == 1
            vir_idx = mo_occ == 0
            logger.debug(mf, '                  Roothaan           | alpha              | beta')
            logger.debug(mf, '  Highest 2-occ = %18.15g | %18.15g | %18.15g',
                         max(mo_energy[core_idx]),
                         max(mo_ea[core_idx]), max(mo_eb[core_idx]))
            logger.debug(mf, '  Lowest 0-occ =  %18.15g | %18.15g | %18.15g',
                         min(mo_energy[vir_idx]),
                         min(mo_ea[vir_idx]), min(mo_eb[vir_idx]))
            for i in numpy.where(open_idx)[0]:
                logger.debug(mf, '  1-occ =         %18.15g | %18.15g | %18.15g',
                             mo_energy[i], mo_ea[i], mo_eb[i])

        if mf.verbose >= logger.DEBUG:
            numpy.set_printoptions(threshold=nmo)
            logger.debug(mf, '  Roothaan mo_energy =\n%s', mo_energy)
            logger.debug1(mf, '  alpha mo_energy =\n%s', mo_ea)
            logger.debug1(mf, '  beta  mo_energy =\n%s', mo_eb)
            numpy.set_printoptions(threshold=1000)
    return mo_occ

def _fill_rohf_occ(mo_energy, mo_energy_a, mo_energy_b, ncore, nopen):
    mo_occ = numpy.zeros_like(mo_energy)
    open_idx = []
    core_sort = numpy.argsort(mo_energy)
    core_idx = core_sort[:ncore]
    if nopen > 0:
        open_idx = core_sort[ncore:]
        open_sort = numpy.argsort(mo_energy_a[open_idx])
        open_idx = open_idx[open_sort[:nopen]]
    mo_occ[core_idx] = 2
    mo_occ[open_idx] = 1
    return mo_occ

def get_grad(mo_coeff, mo_occ, fock):
    '''ROHF gradients is the off-diagonal block [co + cv + ov], where
    [ cc co cv ]
    [ oc oo ov ]
    [ vc vo vv ]
    '''
    occidxa = mo_occ > 0
    occidxb = mo_occ == 2
    viridxa = ~occidxa
    viridxb = ~occidxb
    uniq_var_a = viridxa.reshape(-1,1) & occidxa
    uniq_var_b = viridxb.reshape(-1,1) & occidxb

    if getattr(fock, 'focka', None) is not None:
        focka = fock.focka
        fockb = fock.fockb
    elif isinstance(fock, (tuple, list)) or getattr(fock, 'ndim', None) == 3:
        focka, fockb = fock
    else:
        focka = fockb = fock
    focka = mo_coeff.conj().T.dot(focka).dot(mo_coeff)
    fockb = mo_coeff.conj().T.dot(fockb).dot(mo_coeff)

    g = numpy.zeros_like(focka)
    g[uniq_var_a]  = focka[uniq_var_a]
    g[uniq_var_b] += fockb[uniq_var_b]
    return g[uniq_var_a | uniq_var_b]

def make_rdm1(mo_coeff, mo_occ, **kwargs):
    '''One-particle density matrix.  mo_occ is a 1D array, with occupancy 1 or 2.
    '''
    if isinstance(mo_occ, numpy.ndarray) and mo_occ.ndim == 1:
        mo_occa = (mo_occ > 0).astype(numpy.double)
        mo_occb = (mo_occ ==2).astype(numpy.double)
    else:
        mo_occa, mo_occb = mo_occ
    dm_a = numpy.dot(mo_coeff*mo_occa, mo_coeff.conj().T)
    dm_b = numpy.dot(mo_coeff*mo_occb, mo_coeff.conj().T)
    return lib.tag_array((dm_a, dm_b), mo_coeff=mo_coeff, mo_occ=mo_occ)

def energy_elec(mf, dm=None, h1e=None, vhf=None):
    if dm is None: dm = mf.make_rdm1()
    elif isinstance(dm, numpy.ndarray) and dm.ndim == 2:
        dm = numpy.array((dm*.5, dm*.5))
    return uhf.energy_elec(mf, dm, h1e, vhf)

get_veff = uhf.get_veff

def analyze(mf, verbose=logger.DEBUG, with_meta_lowdin=WITH_META_LOWDIN,
            origin=None, **kwargs):
    '''Analyze the given SCF object:  print orbital energies, occupancies;
    print orbital coefficients; Mulliken population analysis
    '''
    from pyscf.lo import orth
    from pyscf.tools import dump_mat
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    mo_coeff = mf.mo_coeff
    log = logger.new_logger(mf, verbose)
    if log.verbose >= logger.NOTE:
        mf.dump_scf_summary(log)

        log.note('**** MO energy ****')
        if getattr(mo_energy, 'mo_ea', None) is not None:
            mo_ea = mo_energy.mo_ea
            mo_eb = mo_energy.mo_eb
            log.note('                Roothaan           | alpha              | beta')
            for i,c in enumerate(mo_occ):
                log.note('MO #%-3d energy= %-18.15g | %-18.15g | %-18.15g occ= %g',
                         i+MO_BASE, mo_energy[i], mo_ea[i], mo_eb[i], c)
        else:
            for i,c in enumerate(mo_occ):
                log.note('MO #%-3d energy= %-18.15g occ= %g',
                         i+MO_BASE, mo_energy[i], c)

    ovlp_ao = mf.get_ovlp()
    if log.verbose >= logger.DEBUG:
        label = mf.mol.ao_labels()
        if with_meta_lowdin:
            log.debug(' ** MO coefficients (expansion on meta-Lowdin AOs) **')
            orth_coeff = orth.orth_ao(mf.mol, 'meta_lowdin', s=ovlp_ao)
            c = reduce(numpy.dot, (orth_coeff.conj().T, ovlp_ao, mo_coeff))
        else:
            log.debug(' ** MO coefficients (expansion on AOs) **')
            c = mo_coeff
        dump_mat.dump_rec(mf.stdout, c, label, start=MO_BASE, **kwargs)
    dm = mf.make_rdm1(mo_coeff, mo_occ)
    if with_meta_lowdin:
        pop_and_charge = mf.mulliken_meta(mf.mol, dm, s=ovlp_ao, verbose=log)
    else:
        pop_and_charge = mf.mulliken_pop(mf.mol, dm, s=ovlp_ao, verbose=log)
    dip = mf.dip_moment(mf.mol, dm, origin=origin, verbose=log)
    return pop_and_charge, dip

mulliken_pop = hf.mulliken_pop
mulliken_meta = hf.mulliken_meta

def canonicalize(mf, mo_coeff, mo_occ, fock=None):
    '''Canonicalization diagonalizes the Fock matrix within occupied, open,
    virtual subspaces separatedly (without change occupancy).
    '''
    if getattr(fock, 'focka', None) is None:
        dm = mf.make_rdm1(mo_coeff, mo_occ)
        fock = mf.get_fock(dm=dm)
    mo_e, mo_coeff = hf.canonicalize(mf, mo_coeff, mo_occ, fock)
    fa, fb = fock.focka, fock.fockb
    mo_ea = numpy.einsum('pi,pi->i', mo_coeff.conj(), fa.dot(mo_coeff)).real
    mo_eb = numpy.einsum('pi,pi->i', mo_coeff.conj(), fb.dot(mo_coeff)).real
    mo_e = lib.tag_array(mo_e, mo_ea=mo_ea, mo_eb=mo_eb)
    return mo_e, mo_coeff

dip_moment = hf.dip_moment


# use UHF init_guess, get_veff, diis, and intermediates such as fock, vhf, dm
# keep mo_energy, mo_coeff, mo_occ as RHF structure

class ROHF(hf.RHF):
    __doc__ = hf.SCF.__doc__

    def __init__(self, mol):
        hf.SCF.__init__(self, mol)
        self.nelec = None

    @property
    def nelec(self):
        if getattr(self, '_nelec', None) is not None:
            return self._nelec
        else:
            return self.mol.nelec
    @nelec.setter
    def nelec(self, x):
        self._nelec = x

    @property
    def nelectron_alpha(self):
        return self.nelec[0]
    @nelectron_alpha.setter
    def nelectron_alpha(self, x):
        logger.warn(self, 'WARN: Attribute .nelectron_alpha is deprecated. '
                    'Set .nelec instead')
        #raise RuntimeError('API updates')
        self.nelec = (x, self.mol.nelectron-x)

    check_sanity = hf.SCF.check_sanity

    def dump_flags(self, verbose=None):
        hf.SCF.dump_flags(self, verbose)
        nelec = self.nelec
        logger.info(self, 'num. doubly occ = %d  num. singly occ = %d',
                    nelec[1], nelec[0]-nelec[1])

    get_init_guess = uhf.UHF.get_init_guess

    def init_guess_by_minao(self, mol=None):
        if mol is None: mol = self.mol
        return init_guess_by_minao(mol)

    def init_guess_by_atom(self, mol=None):
        if mol is None: mol = self.mol
        logger.info(self, 'Initial guess from the superposition of atomic densities.')
        return init_guess_by_atom(mol)

    def init_guess_by_huckel(self, mol=None):
        if mol is None: mol = self.mol
        logger.info(self, 'Initial guess from on-the-fly Huckel, doi:10.1021/acs.jctc.8b01089.')
        return init_guess_by_huckel(mol)

    def init_guess_by_mod_huckel(self, mol=None):
        if mol is None: mol = self.mol
        logger.info(self, '''Initial guess from on-the-fly Huckel, doi:10.1021/acs.jctc.8b01089,
employing the updated GWH rule from doi:10.1021/ja00480a005.''')
        return init_guess_by_mod_huckel(mol)

    def init_guess_by_1e(self, mol=None):
        if mol is None: mol = self.mol
        logger.info(self, 'Initial guess from hcore.')
        h1e = self.get_hcore(mol)
        s1e = self.get_ovlp(mol)
        mo_energy, mo_coeff = self.eig(h1e, s1e)
        mo_occ = self.get_occ(mo_energy, mo_coeff)
        return self.make_rdm1(mo_coeff, mo_occ)

    def init_guess_by_sap(self, mol=None, **kwargs):
        from pyscf.gto.basis import load
        if mol is None: mol = self.mol
        sap_basis = self.sap_basis
        logger.info(self, '''Initial guess from superposition of atomic potentials (doi:10.1021/acs.jctc.8b01089)
This is the Gaussian fit version as described in doi:10.1063/5.0004046.''')
        if isinstance(sap_basis, str):
            atoms = [coord[0] for coord in mol._atom]
            sapbas = {}
            for atom in set(atoms):
                single_element_bs = load(sap_basis, atom)
                if isinstance(single_element_bs, dict):
                    sapbas[atom] = numpy.asarray(single_element_bs[atom][0][1:], dtype=float)
                else:
                    sapbas[atom] = numpy.asarray(single_element_bs[0][1:], dtype=float)
            logger.note(self, f'Found SAP basis {sap_basis.split("/")[-1]}')
        elif isinstance(sap_basis, dict):
            sapbas = {}
            for key in sap_basis:
                sapbas[key] = numpy.asarray(sap_basis[key][0][1:], dtype=float)
        else:
            logger.error(self, 'sap_basis is of an unexpected datatype.')
        return init_guess_by_sap(mol, sap_basis=sapbas, **kwargs)

    def init_guess_by_chkfile(self, chkfile=None, project=None):
        if chkfile is None: chkfile = self.chkfile
        return init_guess_by_chkfile(self.mol, chkfile, project=project)

    get_fock = get_fock
    get_occ = get_occ

    @lib.with_doc(hf.eig.__doc__)
    def eig(self, fock, s):
        e, c = self._eigh(fock, s)
        if getattr(fock, 'focka', None) is not None:
            mo_ea = numpy.einsum('pi,pi->i', c.conj(), fock.focka.dot(c)).real
            mo_eb = numpy.einsum('pi,pi->i', c.conj(), fock.fockb.dot(c)).real
            e = lib.tag_array(e, mo_ea=mo_ea, mo_eb=mo_eb)
        return e, c

    @lib.with_doc(get_grad.__doc__)
    def get_grad(self, mo_coeff, mo_occ, fock=None):
        if fock is None:
            dm1 = self.make_rdm1(mo_coeff, mo_occ)
            fock = self.get_hcore(self.mol) + self.get_veff(self.mol, dm1)
        return get_grad(mo_coeff, mo_occ, fock)

    @lib.with_doc(make_rdm1.__doc__)
    def make_rdm1(self, mo_coeff=None, mo_occ=None, **kwargs):
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if mo_occ is None: mo_occ = self.mo_occ
        if self.mol.spin < 0:
            # Flip occupancies of alpha and beta orbitals
            mo_occ = (numpy.asarray(mo_occ == 2, dtype=numpy.double),
                      numpy.asarray(mo_occ > 0, dtype=numpy.double))
        return make_rdm1(mo_coeff, mo_occ, **kwargs)

    energy_elec = energy_elec

    @lib.with_doc(uhf.get_veff.__doc__)
    def get_veff(self, mol=None, dm=None, dm_last=0, vhf_last=0, hermi=1):
        if mol is None: mol = self.mol
        if dm is None: dm = self.make_rdm1()
        if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
            dm = numpy.array((dm*.5, dm*.5))

        if self._eri is not None or not self.direct_scf:
            if hasattr(dm, 'mo_occ') and numpy.ndim(dm.mo_occ) == 1:
                mo_occa = (dm.mo_occ > 0).astype(numpy.double)
                mo_occb = (dm.mo_occ ==2).astype(numpy.double)
                dm = lib.tag_array(dm, mo_coeff=(dm.mo_coeff,)*2,
                                   mo_occ=(mo_occa,mo_occb))
            vj, vk = self.get_jk(mol, dm, hermi)
            vhf = vj[0] + vj[1] - vk
        else:
            ddm = dm - numpy.asarray(dm_last)
            vj, vk = self.get_jk(mol, ddm, hermi)
            vhf = vj[0] + vj[1] - vk
            vhf += numpy.asarray(vhf_last)
        return vhf

    @lib.with_doc(analyze.__doc__)
    def analyze(self, verbose=None, with_meta_lowdin=WITH_META_LOWDIN,
                **kwargs):
        if verbose is None: verbose = self.verbose
        return analyze(self, verbose, with_meta_lowdin, **kwargs)

    canonicalize = canonicalize

    def spin_square(self, mo_coeff=None, s=None):
        '''Spin square and multiplicity of RHF determinant'''
        neleca, nelecb = self.nelec
        ms = (neleca - nelecb) * .5
        ss = ms * (ms + 1)
        return ss, ms*2+1

    def stability(self,
                  internal=getattr(__config__, 'scf_stability_internal', True),
                  external=getattr(__config__, 'scf_stability_external', False),
                  verbose=None,
                  return_status=False,
                  **kwargs):
        '''
        ROHF/ROKS stability analysis.

        See also pyscf.scf.stability.rohf_stability function.

        Kwargs:
            internal : bool
                Internal stability, within the RHF optimization space.
            external : bool
                External stability. It is not available in current version.
            return_status: bool
                Whether to return `stable_i` and `stable_e`

        Returns:
            If return_status is False (default), the return value includes
            two set of orbitals, which are more close to the stable condition.
            The first corresponds to the internal stability
            and the second corresponds to the external stability.

            Else, another two boolean variables (indicating current status:
            stable or unstable) are returned.
            The first corresponds to the internal stability
            and the second corresponds to the external stability.
        '''
        from pyscf.scf.stability import rohf_stability
        return rohf_stability(self, internal, external, verbose, return_status, **kwargs)

    def nuc_grad_method(self):
        from pyscf.grad import rohf
        return rohf.Gradients(self)

    convert_from_ = hf.RHF.convert_from_

    def to_ks(self, xc='HF'):
        '''Convert to ROKS object.
        '''
        from pyscf import dft
        return self._transfer_attrs_(dft.ROKS(self.mol, xc=xc))

    to_gpu = lib.to_gpu


class HF1e(ROHF):
    scf = hf._hf1e_scf


del (WITH_META_LOWDIN)



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
from copy import deepcopy
from pyscf import ao2mo, fci, mcscf, lib, __config__
from pyscf.lib import logger
from pyscf.dft import gen_grid
from pyscf.mcscf import mc1step
from pyscf.mcscf.addons import StateAverageMCSCFSolver, state_average_mix
from pyscf.mcscf.addons import state_average_mix_, StateAverageMixFCISolver
from pyscf.mcscf.df import _DFCASSCF, _DFCAS
from pyscf.mcpdft import pdft_veff, pdft_feff
from pyscf.mcpdft.otfnal import transfnal, get_transfnal
from pyscf.mcpdft import _dms
from pyscf.mcpdft import chkfile

def energy_tot(mc, mo_coeff=None, ci=None, ot=None, state=0, verbose=None):
    '''Calculate MC-PDFT total energy

    Args:
        mc : an instance of CASSCF or CASCI class
            Note: this function does not currently run the CASSCF or
            CASCI calculation itself prior to calculating the MC-PDFT
            energy. Call mc.kernel () before passing to this function!

    Kwargs:
        mo_coeff : ndarray of shape (nao, nmo)
            Molecular orbital coefficients
        ci : ndarray or list of length (nroots)
            CI vector or vectors.
        ot : an instance of on-top functional class - see otfnal.py
        state : int
            If mc describes a state-averaged calculation, select the
            state (0-indexed).
        verbose : int
            Verbosity of logger output; defaults to mc.verbose

    Returns:
        e_tot : float
            Total MC-PDFT energy including nuclear repulsion energy
        E_ot : float
            On-top (cf. exchange-correlation) energy
    '''
    if ot is None: ot = mc.otfnal
    ot.reset(mol=mc.mol)  # scanner mode safety
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if verbose is None: verbose = mc.verbose
    t0 = (logger.process_clock(), logger.perf_counter())

    # Allow MC-PDFT to be subclassed, and also allow this function to be
    # called without mc being an instance of MC-PDFT class

    casdm1s = mc.make_one_casdm1s(ci, state=state)
    casdm2 = mc.make_one_casdm2(ci, state=state)
    t0 = logger.timer(ot, 'rdms', *t0)

    if callable(getattr(mc, 'energy_mcwfn', None)):
        e_mcwfn = mc.energy_mcwfn(ot=ot, mo_coeff=mo_coeff, casdm1s=casdm1s,
                                  casdm2=casdm2, verbose=verbose)
    else:
        e_mcwfn = energy_mcwfn(ot=ot, mo_coeff=mo_coeff, casdm1s=casdm1s,
                               casdm2=casdm2, verbose=verbose)
    t0 = logger.timer(ot, 'MC wfn energy', *t0)

    if callable(getattr(mc, 'energy_dft', None)):
        e_dft = mc.energy_dft(ot=ot, mo_coeff=mo_coeff, casdm1s=casdm1s,
                              casdm2=casdm2)
    else:
        e_dft = energy_dft(ot=ot, mo_coeff=mo_coeff, casdm1s=casdm1s,
                           casdm2=casdm2)
    t0 = logger.timer(ot, 'E_ot', *t0)

    e_tot = e_mcwfn + e_dft
    return e_tot, e_dft


# Consistency with PySCF convention
kernel = energy_tot  # backwards compatibility


def energy_elec(mc, *args, **kwargs):
    e_tot, E_ot = energy_tot(mc, *args, **kwargs)
    e_elec = e_tot - mc._scf.energy_nuc()
    return e_elec, E_ot


def energy_mcwfn(mc, mo_coeff=None, ci=None, ot=None, state=0, casdm1s=None,
                 casdm2=None, verbose=None):
    '''Compute the parts of the MC-PDFT energy arising from the wave
    function

    Args:
        mc : an instance of CASSCF or CASCI class
            Note: this function does not currently run the CASSCF or
            CASCI calculation itself prior to calculating the MC-PDFT
            energy. Call mc.kernel () before passing to this function!

    Kwargs:
        mo_coeff : ndarray of shape (nao, nmo)
            contains molecular orbital coefficients
        ci : list or ndarray
            contains ci vectors
        ot : an instance of on-top functional class - see otfnal.py
        state : int
            If mc describes a state-averaged calculation, select the
            state (0-indexed).
        casdm1s : ndarray or compatible of shape (2,ncas,ncas)
            Contains spin-separated active-space 1RDM
        casdm2 : ndarray or compatible of shape [ncas,]*4
            Contains spin-summed active-space 2RDM

    Returns:
        e_mcwfn : float
            Energy from the multiconfigurational wave function:
            nuclear repulsion + 1e + coulomb
    '''

    if ot is None: ot = mc.otfnal
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if verbose is None: verbose = mc.verbose
    if casdm1s is None: casdm1s = mc.make_one_casdm1s(ci=ci, state=state)
    if casdm2 is None: casdm2 = mc.make_one_casdm2(ci=ci, state=state)
    log = logger.new_logger(mc, verbose=verbose)
    dm1s = _dms.casdm1s_to_dm1s(mc, casdm1s, mo_coeff=mo_coeff)
    cascm2 = _dms.dm2_cumulant(casdm2, casdm1s)

    hyb_x, hyb_c = ot._numint.rsh_and_hybrid_coeff(ot.otxc, mc.mol.spin)[2]

    Vnn = mc._scf.energy_nuc()
    h = mc._scf.get_hcore()
    dm1 = dm1s[0] + dm1s[1]
    if log.verbose >= logger.DEBUG or abs(hyb_x) > 1e-10:
        vj, vk = mc._scf.get_jk(dm=dm1s)
        vj = vj[0] + vj[1]

    else:
        vj = mc._scf.get_j(dm=dm1)

    Te_Vne = np.tensordot(h, dm1)
    # (vj_a + vj_b) * (dm_a + dm_b)
    E_j = np.tensordot(vj, dm1) / 2
    log.debug('CAS energy decomposition:')
    log.debug('Vnn = %s', Vnn)
    log.debug('Te + Vne = %s', Te_Vne)
    log.debug('E_j = %s', E_j)

    if abs(hyb_x - hyb_c) > 1e-10:
        log.warn("exchange and correlation hybridization differ")
        log.warn("may lead to unphysical results, see https://github.com/pyscf/pyscf-forge/issues/128")

    # Note: this is not the true exchange energy, but just the HF-like exchange
    E_x = 0.0
    if log.verbose >= logger.DEBUG or abs(hyb_x) > 1e-10:
        # (vk_a * dm_a) + (vk_b * dm_b)
        E_x = -(np.tensordot(vk[0], dm1s[0]) + np.tensordot(vk[1], dm1s[1]))
        E_x /= 2.0
        log.debug("E_x = %s", E_x)
        log.debug("Adding (%s) * E_x = %s", hyb_x, hyb_x * E_x)

    # This is not correlation, but the 2-body cumulant tensored with the eri's:
    # g_pqrs * l_pqrs / 2
    E_c = 0.0
    if log.verbose >= logger.DEBUG or abs(hyb_c) > 1e-10:
        aeri = ao2mo.restore(1, mc.get_h2eff(mo_coeff), mc.ncas)
        E_c = np.tensordot(aeri, cascm2, axes=4) / 2
        log.debug("E_c = %s", E_c)
        log.debug("Adding (%s) * E_c = %s", hyb_c, hyb_c * E_c)

    e_mcwfn = Vnn + Te_Vne + E_j + (hyb_x * E_x) + (hyb_c * E_c)
    return e_mcwfn


def energy_dft(mc, mo_coeff=None, ci=None, ot=None, state=0, casdm1s=None,
               casdm2=None, max_memory=None, hermi=1):
    ''' Wrap to ot.energy_ot for subclassing. '''
    if ot is None: ot = mc.otfnal
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if casdm1s is None: casdm1s = mc.make_one_casdm1s(ci, state=state)
    if casdm2 is None: casdm2 = mc.make_one_casdm2(ci, state=state)
    if max_memory is None: max_memory = mc.max_memory
    return ot.energy_ot(casdm1s, casdm2, mo_coeff, mc.ncore,
                        max_memory=max_memory, hermi=hermi)


def get_energy_decomposition(mc, mo_coeff=None, ci=None, ot=None, otxc=None,
                             grids_level=None, grids_attr=None,
                             split_x_c=None, verbose=None):
    '''Compute a decomposition of the MC-PDFT energy into nuclear
    potential (E0), one-electron (E1), Coulomb (E2c), exchange (EOTx),
    correlation (EOTc) terms, and additionally the nonclassical part
    (E2nc) of the MC-SCF energy:

    E(MC-SCF) = E0 + E1 + E2c + Enc
    E(MC-PDFT) = E0 + E1 + E2c + EOTx + EOTc

    For hybrid functionals,

    E(MC-PDFT) = E0 + E1 + E2c + EOTx + EOTc + Enc

    Where the Enc and EOTx/c terms are premultiplied by the hybrid factor. If
    mc.fcisolver.nroots > 1, lists are returned for everything except the
    nuclear potential energy.

    Args:
        mc : an instance of CASSCF or CASCI class

    Kwargs:
        mo_coeff : ndarray
            Contains MO coefficients
        ci : ndarray or list of length nroots
            Contains CI vectors
        ot : an instance of (translated) on-top density fnal class
        otxc : string
            identity of translated functional; overrides ot
        grids_level : integer
            level preset for DFT quadrature grids
        grids_attr : dictionary
            general attributes for DFT quadrature grids
        split_x_c : logical
            whether to split the exchange and correlation parts of the
            ot functional into two separate contributions


    Returns:
        e_nuc : float
            E0 = sum_A>B ZA*ZB/rAB
        e_1e : float or list of length nroots
            E1 = <T+sum_A ZA/rA>
        e_coul : float or list of length nroots
            E2c = 1/2 int rho(1)rho(2)/r12 d1d2
        e_otxc : float or list of length nroots
            EOTxc = translated functional energy
            if split_x_c == True, this is instead
            EOTx = exchange part of translated functional energy
        e_otc : float or list of length nroots
            only returned if split_x_c == True
            EOTc = correlation part of translated functional
        e_ncwfn : float or list of length nroots
            E2ncc = <H> - E0 - E1 - E2c
            If hybrid functional, this term is weighted appropriately. For pure
            functionals, it is the full NC component
    '''
    if verbose is None: verbose = mc.verbose
    log = logger.new_logger(mc, verbose)
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if grids_attr is None: grids_attr = {}
    if grids_level is not None: grids_attr['level'] = grids_level
    if len(grids_attr) or (otxc is not None):
        old_ot = ot if (ot is not None) else mc.otfnal
        old_grids = old_ot.grids
        # TODO: general compatibility with arbitrary (non-translated) fnals
        if otxc is None: otxc = old_ot.otxc
        new_ot = get_transfnal(mc.mol, otxc)
        new_ot.grids.__dict__.update(old_grids.__dict__)
        new_ot.grids.__dict__.update(**grids_attr)
        ot = new_ot
    if ot is None: ot = mc.otfnal
    if split_x_c is None:
        split_x_c = True
        log.warn(
            'Currently, split_x_c in get_energy_decomposition defaults to '
            'True.\nThis default will change to False in the near future.'
        )

    hyb_x, hyb_c = ot._numint.hybrid_coeff(ot.otxc)
    if abs(hyb_x - hyb_c) > 1e-11:
        raise NotImplementedError("hybrid functionals with different exchange, correlations components")
    if not isinstance(ot, transfnal):
        raise NotImplementedError("Decomp for non-translated PDFT fnals")

    if split_x_c:
        ot = list(ot.split_x_c())
    else:
        ot = [ot, ]

    nroots = getattr(mc.fcisolver, 'nroots', 1)

    e_nuc = mc._scf.energy_nuc()

    if nroots > 1:
        e_1e = []
        e_coul = []
        e_otxc = []
        e_ncwfn = []
        for ix in range(nroots):
            row = _get_e_decomp(mc, mo_coeff, ci, ot, state=ix)
            e_1e.append(row[0])
            e_coul.append(row[1])
            e_otxc.append(row[2])
            e_ncwfn.append(row[3])
        e_otxc = [[e[i] for e in e_otxc] for i in range(len(e_otxc[0]))]
    else:
        e_1e, e_coul, e_otxc, e_ncwfn =  _get_e_decomp(mc, mo_coeff, ci, ot)

    if split_x_c:
        e_otx, e_otc = e_otxc
        return e_nuc, e_1e, e_coul, e_otx, e_otc, e_ncwfn
    else:
        return e_nuc, e_1e, e_coul, e_otxc[0], e_ncwfn

def _get_e_decomp(mc, mo_coeff=None, ci=None, ot=None, state=0, verbose=None):
    ncore = mc.ncore
    ncas = mc.ncas
    if ot is None: ot = [mc.otfnal,]
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if verbose is None: verbose = mc.verbose

    if len(ot) == 1:
        hyb_x, hyb_c = ot[0]._numint.hybrid_coeff(ot[0].otxc)

    elif len(ot) == 2:
        hyb_x, hyb_c = [
            fnal._numint.hybrid_coeff(fnal.otxc)[idx] for idx, fnal in enumerate(ot)
        ]

    else:
        raise ValueError("ot must be length of 1 or 2")

    if abs(hyb_x - hyb_c) > 1e-11:
        raise NotImplementedError(
            "hybrid functionals with different exchange, correlations components"
        )

    casdm1s = mc.make_one_casdm1s(ci, state=state)
    casdm1 = casdm1s[0] + casdm1s[1]
    casdm2 = mc.make_one_casdm2(ci, state=state)
    dm1s = _dms.casdm1s_to_dm1s(mc, casdm1s, mo_coeff=mo_coeff,
            ncore=ncore, ncas=ncas)
    dm1 = dm1s[0] + dm1s[1]
    e_nuc = mc._scf.energy_nuc()
    h = mc.get_hcore()
    h1, h0 = mc.h1e_for_cas()
    h2 = ao2mo.restore(1, mc.get_h2eff(), ncas)
    j = mc._scf.get_j(dm=dm1)
    e_1e = np.dot(h.ravel(), dm1.ravel())
    e_coul = np.dot(j.ravel(), dm1.ravel()) / 2

    e_mcscf = h0 + np.dot(h1.ravel(), casdm1.ravel()) + (
            np.dot(h2.ravel(), casdm2.ravel()) * 0.5)
    e_otxc = [fnal.energy_ot(casdm1s, casdm2, mo_coeff, ncore,
                             max_memory=mc.max_memory)
              for fnal in ot]
    e_ncwfn = e_mcscf - e_nuc - e_1e - e_coul

    if abs(hyb_x) > 1e-10:
        e_ncwfn = hyb_x * e_ncwfn

    return e_1e, e_coul, e_otxc, e_ncwfn

class _mcscf_env:
    '''Prevent MC-SCF step of MC-PDFT from overwriting redefined
    quantities e_states and e_tot '''

    def __init__(self, mc):
        self.mc = mc
        self.e_tot = deepcopy(self.mc.e_tot)
        self.e_states = deepcopy(getattr(self.mc, 'e_states', None))

    def __enter__(self):
        self.mc._in_mcscf_env = True

    def __exit__(self, type, value, traceback):
        self.mc.e_tot = self.e_tot
        if getattr(self.mc, 'e_states', None) is not None:
            self.mc.e_mcscf = np.array(self.mc.e_states)
        if self.e_states is not None:
            try:
                self.mc.e_states = self.e_states
            except AttributeError as e:
                self.mc.fcisolver.e_states = self.e_states
                assert (self.mc.e_states is self.e_states), str(e)
            # TODO: redesign this. MC-SCF e_states is stapled to
            # fcisolver.e_states, but I don't want MS-PDFT to be
            # because that makes no sense
        self.mc._in_mcscf_env = False


class _PDFT:
    # Metaclass parent; unusable on its own
    '''MC-PDFT child class. All total energy quantities (e_tot,
    e_states) are MC-PDFT total energies:

    E = Vnn + h_pq*D_pq + g_pqrs*D_pq*D_rs/2 + E_ot
      = T + Vnn + Vne + E_Coul + E_ot
      = E_classical + E_ot

    Extra attributes:
        otfnal : instance of :class:`otfnal`
            The on-top energy functional class
        otxc : string
            Synonym for `otfnal.otxc`
        grids : instance of :class:`Grids`
            Synonym for `otfnal.grids`

    Additional saved results:
        e_mcscf : float or list of length nroots
            MC-SCF total energy or energies:
            Vnn + h_pq*D_pq + g_pqrs*d_pqrs/2
        e_ot : float or list of length nroots
            On-top nonclassical term in the MC-PDFT energy
    '''

    _mc_class = None

    def __init__(self, scf, ncas, nelecas, my_ot=None, grids_level=None,
                 grids_attr=None, **kwargs):
        # Keep the same initialization pattern for backwards-compatibility.
        # Use a separate intializer for the ot functional
        if grids_attr is None: grids_attr = {}
        _mc_class_no_df = self._mc_class
        if issubclass (_mc_class_no_df, _DFCAS):
            _mc_class_no_df = lib.drop_class (_mc_class_no_df, _DFCAS)
        _mc_class_no_df.__init__(self, scf, ncas, nelecas)
        if issubclass (self._mc_class, _DFCAS):
            self._mc_class.__init__(self, self, scf.with_df)
        keys = set(('e_ot', 'e_mcscf', 'get_pdft_veff', 'get_pdft_feff', 'e_states', 'otfnal',
                    'grids', 'max_cycle_fp', 'conv_tol_ci_fp', 'mcscf_kernel', 'chkfile'))
        self.max_cycle_fp = getattr(__config__, 'mcscf_mcpdft_max_cycle_fp',
                                    50)
        self.conv_tol_ci_fp = getattr(__config__,
                                      'mcscf_mcpdft_conv_tol_ci_fp', 1e-8)
        self.mcscf_kernel = self._mc_class.kernel
        self.chkfile = self._scf.chkfile
        self._in_mcscf_env = False
        self._keys = set(self.__dict__.keys()).union(keys)
        if grids_level is not None:
            grids_attr['level'] = grids_level
        if my_ot is not None:
            self._init_ot_grids(my_ot, grids_attr=grids_attr)

    def _init_ot_grids(self, my_ot, grids_attr=None):
        if grids_attr is None: grids_attr = {}
        old_grids = getattr(self, 'grids', None)
        if isinstance(my_ot, (str, np.bytes_)):
            self.otfnal = get_transfnal(self.mol, my_ot)
        else:
            self.otfnal = my_ot
        if isinstance(old_grids, gen_grid.Grids):
            self.otfnal.grids = old_grids
        # self.grids = self.otfnal.grids
        self.grids.__dict__.update(grids_attr)
        for key in grids_attr:
            assert (getattr(self.grids, key, None) == getattr(
                self.otfnal.grids, key, None))
        # Make sure verbose and stdout don't accidentally change
        # (i.e., in scanner mode)
        self.otfnal.verbose = self.verbose
        self.otfnal.stdout = self.stdout

    def get_rhf_base (self):
        from pyscf.scf.hf import RHF
        from pyscf.scf.rohf import ROHF
        from pyscf.scf.hf_symm import SymAdaptedRHF
        from pyscf.scf.hf_symm import SymAdaptedROHF
        rhf_cls = self._scf.__class__
        if issubclass (rhf_cls, SymAdaptedROHF):
            rhf_cls = lib.replace_class (rhf_cls, SymAdaptedROHF, SymAdaptedRHF)
        if issubclass (rhf_cls, ROHF):
            rhf_cls = lib.replace_class (rhf_cls, ROHF, RHF)
        return lib.view (self._scf, rhf_cls)

    @property
    def grids(self):
        return self.otfnal.grids

    @grids.setter
    def grids(self, x):
        self.otfnal.grids = x
        return self.otfnal.grids

    def optimize_mcscf_(self, mo_coeff=None, ci0=None, **kwargs):
        '''Optimize the MC-SCF wave function underlying an MC-PDFT calculation.
        Has the same calling signature as the parent kernel method. '''
        with _mcscf_env(self):
            self.e_mcscf, self.e_cas, self.ci, self.mo_coeff, self.mo_energy = \
                self._mc_class.kernel(self, mo_coeff, ci0=ci0, **kwargs)
        return self.e_mcscf, self.e_cas, self.ci, self.mo_coeff, self.mo_energy

    def compute_pdft_energy_(self, mo_coeff=None, ci=None, ot=None, otxc=None,
                             grids_level=None, grids_attr=None, dump_chk=True, verbose=None, **kwargs):
        '''Compute the MC-PDFT energy(ies) (and update stored data)
        with the MC-SCF wave function fixed. '''
        if mo_coeff is not None: self.mo_coeff = mo_coeff
        if ci is not None: self.ci = ci
        if ot is not None: self.otfnal = ot
        if otxc is not None: self.otxc = otxc
        if grids_attr is None: grids_attr = {}
        if grids_level is not None: grids_attr['level'] = grids_level
        if len(grids_attr): self.grids.__dict__.update(**grids_attr)
        if verbose is None: verbose = self.verbose
        self.verbose = self.otfnal.verbose = verbose
        nroots = getattr(self.fcisolver, 'nroots', 1)
        epdft = [self.energy_tot(mo_coeff=self.mo_coeff, ci=self.ci, state=ix,
                                 logger_tag='MC-PDFT state {}'.format(ix))
                 for ix in range(nroots)]
        self.e_ot = [e_ot for e_tot, e_ot in epdft]
        if isinstance(self, StateAverageMCSCFSolver):
            e_states = [e_tot for e_tot, e_ot in epdft]
            try:
                self.e_states = e_states
            except AttributeError as e:
                self.fcisolver.e_states = e_states
                assert (self.e_states is e_states), str(e)
            # TODO: redesign this. MC-SCF e_states is stapled to
            # fcisolver.e_states, but I don't want MS-PDFT to be
            # because that makes no sense
            self.e_tot = np.dot(e_states, self.weights)
            e_states = self.e_states
        elif nroots > 1:  # nroots>1 CASCI
            self.e_tot = [e_tot for e_tot, e_ot in epdft]
            e_states = self.e_tot
        else:  # nroots==1 not StateAverage class
            self.e_tot, self.e_ot = epdft[0]
            e_states = [self.e_tot]

        if dump_chk:
            e_tot = self.e_tot
            e_ot = self.e_ot
            self.dump_chk(locals())

        return self.e_tot, self.e_ot, e_states

    def kernel(self, mo_coeff=None, ci0=None, otxc=None, grids_attr=None,
               grids_level=None, **kwargs):
        self.optimize_mcscf_(mo_coeff=mo_coeff, ci0=ci0, **kwargs)
        self.compute_pdft_energy_(otxc=otxc, grids_attr=grids_attr,
                                  grids_level=grids_level, **kwargs)
        # TODO: edit StateAverageMCSCF._finalize in pyscf.mcscf.addons
        # to use the proper name of the class rather than "CASCI", so
        # that I can meaningfully play with "finalize" here
        return (self.e_tot, self.e_ot, self.e_mcscf, self.e_cas, self.ci,
                self.mo_coeff, self.mo_energy)

    def dump_flags(self, verbose=None):
        self._mc_class.dump_flags(self, verbose=verbose)
        log = logger.new_logger(self, verbose)
        log.info('on-top pair density exchange-correlation functional: %s',
                 self.otfnal.otxc)

    def get_pdft_veff(self, mo=None, ci=None, state=0, casdm1s=None,
                      casdm2=None, incl_coul=False, paaa_only=False, aaaa_only=False,
                      jk_pc=False, drop_mcwfn=False, incl_energy=False, ot=None):
        '''Get the 1- and 2-body MC-PDFT effective potentials for a set
        of mos and ci vectors

        Kwargs:
            mo : ndarray of shape (nao,nmo)
                A full set of molecular orbital coefficients. Taken from
                self if not provided
            ci : list or ndarray
                CI vectors. Taken from self if not provided
            state : integer
                Indexes a specific state in state-averaged calculations.
            casdm1s : ndarray of shape (2,ncas,ncas)
                Spin-separated 1-RDM in the active space
            casdm2 : ndarray of shape (ncas,ncas,ncas,ncas)
                Spin-summed 2-RDM in the active space
            incl_coul : logical
                If true, includes the Coulomb repulsion energy in the
                1-body effective potential.
            paaa_only : logical
                If true, only the paaa 2-body effective potential
                elements are evaluated; the rest of ppaa are filled with
                zeros.
            aaaa_only : logical
                If true, only the aaaa 2-body effective potential
                elements are evaluated; the rest of ppaa are filled with
                zeros.
            jk_pc : logical
                If true, calculate the ppii=pipi 2-body effective
                potential in veff2.j_pc and veff2.k_pc. Otherwise these
                arrays are filled with zeroes.
            drop_mcwfn : logical
                If true, drops the normal CASSCF wave function contribution
                (ie the ``Hartree exchange-correlation'') from the response
            incl_energy : logical
                If true, includes the on-top potential energy as a 3rd return argument
            ot : an instance of otfnal class

        Returns:
            veff1 : ndarray of shape (nao, nao)
                1-body effective potential in the AO basis
                May include classical Coulomb potential term (see
                incl_coul kwarg)
            veff2 : pyscf.mcscf.mc_ao2mo._ERIS instance
                Relevant 2-body effective potential in the MO basis
            E_ot : float
                On-top energy. Only included if incl_energy is true.
        '''
        t0 = (logger.process_clock(), logger.perf_counter())
        if mo is None:
            mo = self.mo_coeff
        if ci is None:
            ci = self.ci
        if casdm1s is None:
            casdm1s = self.make_one_casdm1s(ci, state=state)
        if casdm2 is None:
            casdm2 = self.make_one_casdm2(ci, state=state)
        if ot is None:
            ot = self.otfnal

        ncore, ncas = self.ncore, self.ncas
        dm1s = _dms.casdm1s_to_dm1s(self, casdm1s, mo_coeff=mo,
                                    ncore=ncore, ncas=ncas)
        cascm2 = _dms.dm2_cumulant(casdm2, casdm1s)

        spin = abs(self.nelecas[0] - self.nelecas[1])
        omega, _, hyb = ot._numint.rsh_and_hybrid_coeff(ot.otxc, spin=spin)

        if abs(omega) > 1e-11:
            raise NotImplementedError("range-separated on-top 1e potentials")
        if abs(hyb[0] > hyb[1]) > 1e-11:
            raise NotImplementedError("hybrid functionals with different exchange, correlation components")

        cas_hyb = hyb[0]
        ot_hyb = 1.0-cas_hyb

        E_ot, pdft_veff1, pdft_veff2 = pdft_veff.kernel(
            ot,
            dm1s,
            cascm2,
            mo,
            ncore,
            ncas,
            max_memory=self.max_memory,
            paaa_only=paaa_only,
            aaaa_only=aaaa_only,
            jk_pc=jk_pc,
        )

        if incl_coul:
            pdft_veff1 += ot_hyb*self._scf.get_j(self.mol, dm1s[0] + dm1s[1])

        if not drop_mcwfn and cas_hyb > 1e-11:
            raise NotImplementedError("adding mcwfn response to pdft_veff2")

        logger.timer(self, 'get_pdft_veff', *t0)

        if incl_energy:
            return pdft_veff1, pdft_veff2, E_ot

        else:
            return pdft_veff1, pdft_veff2

    def get_pdft_feff(self, mo=None, ci=None, state=0, casdm1s=None,
                      casdm2=None, c_dm1s=None, c_cascm2=None,
                      paaa_only=False, aaaa_only=False, jk_pc=False, incl_coul=False, delta=False):
        """casdm1s and casdm2 are the values that are put into the kernel
        whereas the c_dm1s and c_cascm2 are the densities which multiply the
        kernel function (ie the contraction in terms of normal 1 and 2-rdm
        quantities.)

        incl_coul includes the coulomb interaction with the contracting density!
        delta actually sets contracted density to contracted_density - density (like delta in lpdft grads)
        """
        t0 = (logger.process_clock(), logger.perf_counter())
        if mo is None: mo = self.mo_coeff
        if ci is None: ci = self.ci
        if casdm1s is None: casdm1s = self.make_one_casdm1s(ci, state=state)
        if casdm2 is None: casdm2 = self.make_one_casdm2(ci, state=state)
        ncore, ncas = self.ncore, self.ncas

        dm1s = _dms.casdm1s_to_dm1s(self, casdm1s, mo_coeff=mo, ncore=ncore,
                                    ncas=ncas)
        cascm2 = _dms.dm2_cumulant(casdm2, casdm1s)

        if c_dm1s is None:
            c_dm1s = dm1s

        if c_cascm2 is None:
            c_cascm2 = cascm2

        pdft_feff1, pdft_feff2 = pdft_feff.kernel(self.otfnal, dm1s, cascm2,
                                                  c_dm1s, c_cascm2, mo, ncore,
                                                  ncas,
                                                  max_memory=self.max_memory,
                                                  paaa_only=paaa_only,
                                                  aaaa_only=aaaa_only,
                                                  jk_pc=jk_pc, delta=delta)

        if incl_coul:
            if delta:
                c_dm1s -= dm1s

            pdft_feff1 += self._scf.get_j(self.mol, c_dm1s[0] + c_dm1s[1])

        logger.timer(self, 'get_pdft_feff', *t0)
        return pdft_feff1, pdft_feff2

    def _state_average_nuc_grad_method(self, state=None):
        if not isinstance(self, mc1step.CASSCF):
            raise NotImplementedError("CASCI-based PDFT nuclear gradients")
        elif getattr(self, 'frozen', None) is not None:
            raise NotImplementedError("PDFT nuclear gradients with frozen orbitals")
        elif isinstance(self, _DFCASSCF):
            from pyscf.df.grad.mcpdft import Gradients
        else:
            from pyscf.grad.mcpdft import Gradients
        return Gradients(self, state=state)

    def nuc_grad_method(self):
        return self._state_average_nuc_grad_method(state=None)

    Gradients=nuc_grad_method

    def dip_moment(self, unit='Debye', origin='Coord_Center', state=0):
        if not isinstance(self, mc1step.CASSCF):
            raise NotImplementedError("CASCI-based PDFT dipole moments")
        elif getattr(self, 'frozen', None) is not None:
            raise NotImplementedError("PDFT dipole moments with frozen orbitals")
        elif isinstance(self, _DFCASSCF):
            raise NotImplementedError("PDFT dipole moments with density-fitting ERIs")
        # Monkeypatch for double prop folders
        # TODO: more elegant solution
        import os
        mypath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        myproppath = os.path.join(mypath, 'prop')
        # suppress irrelevant warnings when 'properties' ext mod installed
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="Module.*is under testing")
            from pyscf import prop
        prop.__path__.append(myproppath)
        prop.__path__ = list(set(prop.__path__))
        from pyscf.prop.dip_moment.mcpdft import ElectricDipole
        dip_obj = ElectricDipole(self)
        mol_dipole = dip_obj.kernel(state=state, unit=unit, origin=origin)
        return mol_dipole

    def get_energy_decomposition(self, mo_coeff=None, ci=None, ot=None,
                                 otxc=None, grids_level=None,
                                 grids_attr=None, split_x_c=None,
                                 verbose=None):
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if ci is None: ci = self.ci
        if verbose is None: verbose = self.verbose
        return get_energy_decomposition(
            self, mo_coeff=mo_coeff, ci=ci, ot=ot, otxc=otxc,
            grids_level=grids_level, grids_attr=grids_attr,
            split_x_c=split_x_c, verbose=verbose
        )

    def state_average_mix(self, fcisolvers=None, weights=(0.5, 0.5)):
        return state_average_mix(self, fcisolvers, weights)

    def state_average_mix_(self, fcisolvers=None, weights=(0.5, 0.5)):
        state_average_mix_(self, fcisolvers, weights)
        return self

    def multi_state_mix(self, fcisolvers=None, weights=(0.5, 0.5), method='LIN'):
        if method.upper() == "LIN":
            from pyscf.mcpdft.lpdft import linear_multi_state_mix
            return linear_multi_state_mix(self, fcisolvers=fcisolvers, weights=weights)

        else:
            raise NotImplementedError(f"StateAverageMix not available for {method}")

    def multi_state(self, weights=(0.5, 0.5), method='LIN'):
        if method.upper() == "LIN":
            from pyscf.mcpdft.lpdft import linear_multi_state
            return linear_multi_state(self, weights=weights)

        else:
            from pyscf.mcpdft.mspdft import multi_state
            return multi_state(self, weights=weights,
                               diabatization=method)

    def state_interaction(self, weights=(0.5, 0.5), diabatization='CMS'):
        logger.warn(self, ('"state_interaction" for multi-state PDFT is '
                           'deprecated. Use multi_state instead. In the '
                           'future this will raise an error.'))
        return self.multi_state(weights=weights, diabatization=diabatization)

    @property
    def otxc(self):
        return self.otfnal.otxc

    @otxc.setter
    def otxc(self, x):
        self._init_ot_grids(x)

    make_one_casdm1s = _dms.make_one_casdm1s
    make_one_casdm2 = _dms.make_one_casdm2
    energy_mcwfn = energy_mcwfn
    energy_dft = energy_dft

    def energy_tot(self, mo_coeff=None, ci=None, ot=None, state=0,
                   verbose=None, otxc=None, grids_level=None, grids_attr=None,
                   logger_tag='MC-PDFT'):
        ''' Compute the MC-PDFT energy of a single state '''
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if ci is None: ci = self.ci
        if grids_attr is None: grids_attr = {}
        if grids_level is not None: grids_attr['level'] = grids_level
        if len(grids_attr) or (otxc is not None):
            old_ot = ot if (ot is not None) else self.otfnal
            old_grids = old_ot.grids
            # TODO: general compatibility with arbitrary (non-translated) fnals
            if otxc is None: otxc = old_ot.otxc
            new_ot = get_transfnal(self.mol, otxc)
            new_ot.grids.__dict__.update(old_grids.__dict__)
            new_ot.grids.__dict__.update(**grids_attr)
            ot = new_ot
        elif ot is None:
            ot = self.otfnal
        e_tot, e_ot = energy_tot(self, mo_coeff=mo_coeff, ot=ot, ci=ci,
                                 state=state, verbose=verbose)
        logger.note(self, '%s E = %s, Eot(%s) = %s', logger_tag,
                    e_tot, ot.otxc, e_ot)
        return e_tot, e_ot

    def dump_chk(self, envs):
        """
        Dumps information to the chkfile. If called within mcscf environment,
        it forwards to the mcscf dump_chk. Else, it dumps only the pdft
        information.
        """
        if not self.chkfile:
            return self

        # Hack, basically if we are optimizing mcscf, then call that dump
        # Otherwise, we need to dump the pdft dump...
        if self._in_mcscf_env:
            self._mc_class.dump_chk(self, envs)

        else:
            e_states = None
            if len(envs["e_states"]) > 1:
                e_states = envs["e_states"]

            chkfile.dump_mcpdft(
                self,
                chkfile=self.chkfile,
                key="pdft",
                e_tot=envs["e_tot"],
                e_ot=envs["e_ot"],
                e_states=e_states,
                e_mcscf=self.e_mcscf,
            )

        return self

    def update_from_chk(self, chkfile=None, pdft_key="pdft"):
        if chkfile is None:
            chkfile = self.chkfile

        # When the chkfile is saved, we utilize hard links to the mcscf data
        self.__dict__.update(lib.chkfile.load(chkfile, pdft_key))
        return self

    update = update_from_chk

def get_mcpdft_child_class(mc, ot, **kwargs):
    # Inheritance magic
    mc_doc = (mc.__class__.__doc__ or
              'No docstring for MC-SCF parent method')

    class PDFT(_PDFT, mc.__class__):
        __doc__ = mc_doc + '\n\n' + _PDFT.__doc__
        _mc_class = mc.__class__

    pdft = PDFT(mc._scf, mc.ncas, mc.nelecas, my_ot=ot, **kwargs)
    _keys = pdft._keys.copy()
    pdft.__dict__.update(mc.__dict__)
    pdft._keys = pdft._keys.union(_keys)
    return pdft



#!/usr/bin/env python
# Copyright 2019-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Elvira R. Sayfutyarova
#


''' When using results of this code for publications, please cite the following paper:
    "Constructing molecular pi-orbital active spaces for multireference calculations of conjugated systems"
     E. R. Sayfutyarova and S. Hammes-Schiffer, J. Chem. Theory Comput., 15, 1679 (2019).
'''


import numpy as np
import numpy.linalg as la
from pyscf import gto

def mdot(*args):
    """chained matrix product: mdot(A,B,C,..) = A*B*C*...
    No attempt is made to optimize the contraction order."""
    r = args[0]
    for a in args[1:]:
        r = np.dot(r,a)
    return r

def MakeShellsForElement(mol, Element):
    """make a list with MINAO basis set data for a given element in the mol object"""
    PyScfShellInfo = gto.basis.load(mol.basis, Element)
    AtShells = []
    for ShellData in PyScfShellInfo:
        # see: gto/basis/minao.py on format of data.
        # data comes as a list of:
        #   [l, (...), (...), (...)]
        # where each (...) specifies both exponents and contractions and has the same length.

        l, ExponentsAndContractions = ShellData[0], np.asarray(ShellData[1:])
        # primitive exponents
        # Exp = ExponentsAndContractions[:,0]
        # contraction coefficient matrix
        CGTO = ExponentsAndContractions[:,1:]
        nCGTO = CGTO.shape[1]
        nLk = nCGTO*(2*l+1)
        AtShells.append((l, nCGTO, nLk))
    return AtShells

def MakeShells(mol, Elements):
    """collect MINAO basis set data for all elements"""
    Shells = []
    for iAt in range(mol.natm):
        ElementShells = MakeShellsForElement(mol, Elements[iAt])
        Shells.append(ElementShells)
    return Shells

def Atoms_w_Coords(mol):
    """collect info about atoms' positions"""
    AtomCoords = mol.atom_coords()

    assert (AtomCoords.shape == (mol.natm, 3))
    Elements = [mol.atom_pure_symbol(iAt) for iAt in range(mol.natm)]
    Elements =np.asarray(Elements)

    return Elements,AtomCoords

def MakePiOS(mol,mf,PiAtomsList, nPiOcc=None,nPiVirt=None):
    np.set_printoptions(precision=4,linewidth=10000,edgeitems=3,suppress=False)

    print("================================CONSTRUCTING PI-ORBITALS===========================================")
    # PiAtoms list contains the set of main-group atoms involved in the pi-system of the chromophores(?)
    # indices are 1-based.

    mol2=mol.copy()
    mol2.basis = 'MINAO'
    mol2.build()

    # make a minimal AO basis for our atoms. We load the basis from a library
    # just to have access to its shell-composition. Need that to find the indices
    # of all atoms, and the AO indices of the px/py/pz functions.

    Elements,Coords = Atoms_w_Coords(mol2)
    Shells = MakeShells(mol2,Elements)
    def AssignTag(iAt, Element, Tag):
        assert (Elements[iAt-1] == Element)
        Elements[iAt-1] = Element + Tag


    # fix type of atom for the donor-acceptor which are exchanging the hydrogens.
    # During the process, the hydrogens are moving and the Huckel-Theory
    # formal number of contributed pi-electrons may change.
    # These settings override the auto-detection of number of pi electrons
    # based on atomic connectivity in GetNumPiElec() below.
#    AssignTag(50, "N", "1e")



    # OrbBasis = mol.basis
    C = mf.mo_coeff
    S1 = mol.intor_symmetric("int1e_ovlp")
    S2 = mol2.intor_symmetric("int1e_ovlp")
    S12 = gto.intor_cross('int1e_ovlp', mol, mol2)
    SMo = mdot(C.T, S1, C)
    print("    MO deviation from orthogonality  {:8.2e} \n".format(rmsd(SMo - np.eye(SMo.shape[0]))))

    # make arrays of occupation numbers and orbital eigenvalues.
    Occ = mf.mo_occ
    Eps = mf.mo_energy

    nOrb = C.shape[1]
    if (mol.spin==0):
        nOcc = np.sum(Occ == 2)
    else:
        nOcc = np.sum(Occ == 2)+np.sum(Occ == 1)
        n1=np.sum(Occ == 1)
        print("    Number of singly occupied orbitals      {} ".format(n1))

    nVir = np.sum(Occ == 0)
    assert (nOcc + nVir == nOrb)
    print("    Number of occupied orbitals      {} ".format(nOcc))
    print("    Number of unoccupied orbitals    {} ".format(nVir))

    # Compute Fock matrix from orbital eigenvalues (SCF has to be fully converged)
    Fock = mdot(S1, C, np.diag(Eps), C.T, S1.T)
    # Rdm = mdot(C, np.diag(Occ), C.T)

    COcc = C[:,:nOcc]
    CVir = C[:,nOcc:]

    # Smh1 = MakeSmh(S1)
    # Sh1 = np.dot(S1, Smh1)

    # Compute IAO basis (non-orthogonal). Will be used to identify the
    # pi-MO-space of the target atom groups.
    CIb = MakeIaosRaw(COcc, S1, S2, S12)
    # CIbOcc = np.linalg.lstsq(CIb, COcc)[0]
    # Err = np.dot(Sh1, np.dot(CIb, CIbOcc) - COcc)

    # check orthogonality of IAO basis occupied orbitals
    SIb = mdot(CIb.T, S1, CIb)
    # SIbOcc = mdot(CIbOcc.T, SIb, CIbOcc)
    nIb = SIb.shape[0]

    if 0:
        # make the a representation of the virtual valence space.
        # SmhIb = MakeSmh(SIb)

        nIbVir = nIb - nOcc  # number of virtual valence orbitals

        CTargetIb = CIb
        STargetIb = mdot(CTargetIb.T, S1, CTargetIb)
        SmhTargetIb = MakeSmh(STargetIb)
        STargetIbVir = mdot(SmhTargetIb, CTargetIb.T, S1, CVir)
        U,sig,Vt = np.linalg.svd(STargetIbVir, full_matrices=False)
        print("    Number of target MINAO basis fn     {}\n".format(CTargetIb.shape[1]))

        print("SIbVir Singular Values (n={})".format(len(sig)))
        print("    [{}]".format(', '.join('{:.4f}'.format(k) for k in sig)))


        assert (np.abs(sig[nIbVir-1] - 1.0) < 1e-4)
        assert (np.abs(sig[nIbVir] - 0.0) < 1e-4)

        # for pi systems: do it like here -^ for the virtuals, but
        # for COcc and CVir both. What we should do is:
        #   - instead of CIb, we use CTargetIb = CIb * (AO-linear-comb-matrix)
        #   - instead of SmhIb, we use its corresponding target-AO overlap matrix:
        #     SmhTargetIb = MakeSmh(mdot(CTargetIb.T, S1, CTargetIb))
        #   - instead of using nOcc/nVir/nIb to determine the target number
        #     of orbitals, we obtain the target number of pi electrons by counting
        #     their subset from the selected main group atoms (see CHEM 408 u11).
        #     From this determine the number of occupied pi-orbitals this system
        #     is supposed to have, and virtual orbitals as nTargetIb - nPiOcc
        # The AoMix matrix is made from the pi-system's inertial tensor to get the
        # local z-direction, and then linearly-combining this onto the highest-N p-AOs.

    CActOcc = []
    CActVir = []

    # add pi-HOMOs and pi-LUMOs
    CFragOcc, CFragVir,nOccOrbExpected,nVirtOrbExpected = MakePiSystemOrbitals(
        "Pi-System", PiAtomsList, None, Elements,Coords, CIb, Shells, S1, S12, S2, Fock, COcc, CVir)
    if (nPiOcc is None):
        for i in range(1,nOccOrbExpected+1):
            CActOcc.append(CFragOcc[:,-i])
    else:
        for i in range(1,nPiOcc+1):
            CActOcc.append(CFragOcc[:,-i])

    if (nPiVirt is None):
        for j in range(nVirtOrbExpected):
            CActVir.append(CFragVir[:,j])
    else:
        for j in range(nPiVirt):
            CActVir.append(CFragVir[:,j])

    print("\n -- Joining active spaces")
    if (mol.spin==0):
        nElec = 2*len(CActOcc)
    else:
        nElec = 2*len(CActOcc)-n1
    CAct = np.array(CActOcc + CActVir).T
    if 0:
        # orthogonalize and semi-canonicalize
        SAct = mdot(CAct.T, S1, CAct)
        ew, ev = np.linalg.eigh(SAct)
        print("    CAct initial overlap (if all ~approx 1, then the initial "
              "active orbitals are near-orthogonal. That is good.)")
        print("    [{}]".format(', '.join('{:.4f}'.format(k) for k in ew)))

        CAct = np.dot(CAct, MakeSmh(SAct))
        CAct = SemiCanonicalize(CAct, Fock, S1, "active")

    print("    Number of Active Electrons     {} ".format(nElec))
    print("    Number of Active Orbitals      {} ".format( CAct.shape[1]))

    # make new non-active occupied and non-active virtual orbitals,
    # in order to rebuild a full MO matrix.
    def MakeInactiveSpace(Name, CActList, COrb1):
        CAct = np.array(CActList).T
        SAct = mdot(CAct.T, S1, CAct)
        CAct = np.dot(CAct, MakeSmh(SAct))

        SActMo = mdot(COrb1.T, S1.T, CAct, CAct.T, S1, COrb1)
        ew, ev = np.linalg.eigh(SActMo) # small ews first (orbs not overlapping with active orbs).

        nRest = COrb1.shape[1] - CAct.shape[1]
        if 0:
            for i in range(len(ew)):
                print("{:4} {:15.6f}  {}".format(i,ew[i], i == nRest))
        assert (np.abs(ew[nRest-1] - 0.0) < 1e-8)
        assert (np.abs(ew[nRest] - 1.0) < 1e-8)
        CNewOrb1 = np.dot(COrb1, ev[:,:nRest])
        return SemiCanonicalize(CNewOrb1, Fock, S1, Name, Print=False)

    CNewClo = MakeInactiveSpace("NewClo", CActOcc, COcc)
    CNewExt = MakeInactiveSpace("NewVir", CActVir, CVir)

    # re-orthogonalize (should be orthogonal already, but just to be sure).
    COrbNew = np.hstack([CAct])
    if 1:
        COrbNew = np.hstack([CNewClo, CAct, CNewExt])
        SMo = mdot(COrbNew.T, S1, COrbNew)
        COrbNew = np.dot(COrbNew, MakeSmh(SMo))

    print("    Number of Core Orbitals        {} ".format(CNewClo.shape[1]))
    print("    Number of Virtual Orbitals     {} ".format(CNewExt.shape[1]))
    print("    Total Number of Orbitals       {} ".format(COrbNew.shape[1]))

    return CNewClo.shape[1],CAct.shape[1],CNewExt.shape[1],nElec, COrbNew


def rmsd(a, b = None):
    if b is None:
        return np.mean(a.flatten()**2)**.5
    else:
        return rmsd(a - b)


def MakeSmh(S):
    ew,ev = np.linalg.eigh(S)
    assert (np.all(ew > 1e-10))
    v = ev * (ew**-0.25)[np.newaxis,:]
    return np.dot(v, v.T)



def MakeIaosRaw(COcc, S1, S2, S12):
    # calculate the molecule-intrinsic atomic orbital (IAO) basis
    # ref: [1] Knizia, J. Chem. Theory Comput., http://dx.doi.org/10.1021/ct400687b
    # This is the "Simple/2014" version from ibo-ref at sites.psu.edu/knizia/software
    assert (S1.shape[0] == S1.shape[1] and S1.shape[0] == S12.shape[0])
    assert (S2.shape[0] == S2.shape[1] and S2.shape[0] == S12.shape[1])
    assert (COcc.shape[0] == S1.shape[0] and COcc.shape[1] <= S2.shape[0])
    P12 = la.solve(S1, S12)   # P12 = S1^{-1} S12
    COcc2 = mdot(S12.T, COcc)              # O(N m^2)
    CTil = la.solve(S2, COcc2)             # O(m^3)
    STil = mdot(COcc2.T, CTil)             # O(m^3)
    CTil2Bar = la.solve(STil, CTil.T).T    # O(m^3)
    T4 = COcc - mdot(P12, CTil2Bar)        # O(N m^2)
    CIb = P12 + mdot(T4, COcc2.T)          # O(N m^2)
    return CIb





def GetPzOrientation(iTargetAtoms, Coords_, Elements_):
    # make a xyz vector pointing out of the plane containing the target atoms.
    assert (Coords_.shape[1] == 3)
    Coords = Coords_[iTargetAtoms,:].copy()
    Elements = Elements_[iTargetAtoms]

    # get center of mass coordinates
    vCom = np.mean(Coords, axis=0)
    Coords -= vCom

    I = np.zeros((3,3))
    for vCoord in Coords:
        I += np.outer(vCoord, vCoord)
    # diagonalize inertial-like tensor; large eigenvalues first.
    ew,ev = np.linalg.eigh(-I)
    ew *= -1

    print("Target Atom List:")
    print("[{}]".format(', '.join('{}'.format(k) for k in iTargetAtoms)))

    print("Elements        :")
    print("[{}]".format(', '.join('{}'.format(k) for k in Elements)))


    # direction of smallest spatial extend (last one) should be
    # the one indicating the pz-orbital direction.
    vPz = ev[:,2]
    return vPz


def FindValenceAoIndices(iAt, Shells, TargetL):
    ipxyz = None

    iFn0 = 0
    for Atom in range(len(Shells)):
        for AtomL in range(len(Shells[Atom])):
            if Atom == iAt:
                if Shells[Atom][AtomL][0] == TargetL:
                    # this is a generally contracted p-shell on atom iAt.
                    assert (ipxyz is None) # should be only one per atom, I think.
                    nCGTO = Shells[Atom][AtomL][1]
                    nSphComp = 2*TargetL + 1
                    iFnHighestP = iFn0 + nSphComp*(nCGTO-1)
                    ipxyz = np.array([(iFnHighestP + o) for o in range(nSphComp)])
            iFn0 += Shells[Atom][AtomL][2]
    assert (ipxyz is not None)
    return ipxyz



def MakePzMinaoVectors(iTargetAtoms, vPz, Shells):
    # nIb = len(Shells)*len(Shells[Atom])*len( Shells[AtomShell][AtomL])
    nIb=0
    for Atom in range(len(Shells)):
        for AtomL in range(len(Shells[Atom])):
            nIb += Shells[Atom][AtomL][2]
    nVec = len(iTargetAtoms)

    AoVecs = np.zeros((nIb, nVec))
    for (iVec, iAt) in enumerate(iTargetAtoms):
        ipxyz = FindValenceAoIndices(iAt, Shells, TargetL = 1)
        AoVec = np.zeros(nIb)
        AoVec[ipxyz] = vPz

        # each atom brings one pz orbital -> iVec = iiAt
        AoVecs[:,iVec] = AoVec
    return AoVecs


# table of numbers of pi electrons each main group element
# typically brings into a pi-system
_NumPiElec = {
    "C": 1,
    "B": 0,
    "N1e": 1,
    "N2e": 2,
    "O1e": 1,
    "O2e": 2,
    "F": 2,
    "Si":1,
    "P1e": 1,
    "P2e": 2,
    "S1e": 1,
    "S2e": 2,
    "C2e": 2,
}


_CovalentRadii = {1: 38.0, 2: 32.0, 3: 134.0, 4: 90.0, 5: 82.0,
                  6: 77.0, 7: 75.0, 8: 73.0, 9: 71.0, 10: 69.0,
                  11: 154.0, 12: 130.0, 13: 118.0, 14: 111.0, 15: 106.0,
                  16: 102.0, 17: 99.0, 18: 97.0, 19: 196.0, 20: 174.0,
                  21: 144.0, 22: 136.0, 23: 125.0, 24: 127.0, 25: 139.0,
                  26: 125.0, 27: 126.0, 28: 121.0, 29: 138.0, 30: 131.0,
                  31: 126.0, 32: 122.0, 33: 119.0, 34: 116.0, 35: 114.0,
                  36: 110.0, 37: 211.0, 38: 192.0, 39: 162.0, 40: 148.0,
                  41: 137.0, 42: 145.0, 43: 156.0, 44: 126.0, 45: 135.0,
                  46: 131.0, 47: 153.0, 48: 148.0, 49: 144.0, 50: 141.0,
                  51: 138.0, 52: 135.0, 53: 133.0, 54: 130.0, 55: 225.0,
                  56: 198.0, 57: 169.0, 58: None, 59: None, 60: None,
                  61: None, 62: None, 63: None, 64: None, 65: None,
                  66: None, 67: None, 68: None, 69: None, 70: None,
                  71: 160.0, 72: 150.0, 73: 138.0, 74: 146.0, 75: 159.0,
                  76: 128.0, 77: 137.0, 78: 128.0, 79: 144.0, 80: 149.0,
                  81: 148.0, 82: 147.0, 83: 146.0, 84: None, 85: None,
                  86: 145.0, 87: None, 88: None, 89: None, 90: None,
                  91: None, 92: None, 93: None, 94: None, 95: None,
                  96: None, 97: None, 98: None, 99: None, 100: None,
                  101: None, 102: None, 103: None, 104: None, 105: None,
                  106: None, 107: None, 108: None, 109: None, 110: None,
                  111: None, 112: None, 113: None, 114: None, 115: None, 116: None}
# ^- embedded now to remove dependency on finding .txt path.


def GetCovalentRadius(At):
    # get covalent radius in pm
    rcov = _CovalentRadii[At.iElement]
    assert (rcov is not None)
    # convert to bohr radii
    ToAng = 0.52917721092
    return (0.01 * rcov)/ToAng

def GetNumPiElec(iAt, Elements,Coords):
    # deal with C, F, B, Si, and the atoms which have explicit tages assigned.
    At = Elements[iAt]
    nPiTab = _NumPiElec.get(At, None)
    if nPiTab is not None:
        return nPiTab
    print(" ...determining formal number of pi-electrons for {} (not in table).".format(At.Label))

    # find number of bonded partners
    iBonded = []
    for (jAt, AtJ) in enumerate(Elements):
        if iAt == jAt:
            continue
        rij = np.sum((Coords[At] - Coords[AtJ])**2)**.5
        if rij < 1.3 * (GetCovalentRadius(At) + GetCovalentRadius(AtJ)):
            iBonded.append(jAt)

    nBonds = len(iBonded)
    if At in ["N", "P"]:  # 5 valence electrons
        if nBonds == 2:
            # two sigma bonds (1e each) + 1 sigma lone pair (2e) -> 1e left for pi system
            return 1
        elif nBonds == 3:
            # three sigma bonds (1e each) -> 2e left for pi system
            return 2

    if At in ["O","S"]:  # 6 valence electrons
        if nBonds == 1:
            # one sigma bond (1e) + 2 lone pairs in x/y planes (2e each) -> 1e left for pi system
            return 1
        elif nBonds == 3:
            # two sigma bond (1e each) + 1 lone pairs in x/y planes (2e) -> 2e left for pi system
            return 2
    raise Exception("GetNumPiElec({}): {} with {} bonds not tabulated :(".format(iAt, At,  nBonds))


def SemiCanonicalize(COut, Fock, S1, Name, Print=True):
    nOrb = COut.shape[1]
    SOrb = mdot(COut.T, S1, COut)
    fErr = rmsd(SOrb - np.eye(nOrb))
    if fErr > 1e-8:
        print("    SemiCan: Orbital deviation from orthogonality  {:8.2e} \n".format(fErr))
        raise Exception("Orbitals not orthogonal.")
    FockMo = mdot(COut.T, Fock, COut)
    ew,ev = np.linalg.eigh(FockMo)
    if Print:
        print("    Semi-canonicalized orbital energies ({} subspace)".format(Name))
        print("    [{}]".format(', '.join('{:.4f}'.format(k) for k in ew)))
    return np.dot(COut, ev)


def MakeOverlappingOrbSubspace(Space, Name, COrb, nOrbExpected,  CTargetIb, S1, Fock):
    print("\n -- Constructing MO subspace space {}/{}.".format(Space, Name))
    STargetIb = mdot(CTargetIb.T, S1, CTargetIb)
    SmhTargetIb = MakeSmh(STargetIb)

    STargetIbVir = mdot(SmhTargetIb, CTargetIb.T, S1, COrb)
    U,sig,Vt = np.linalg.svd(STargetIbVir, full_matrices=False)

    print("    S[{},{}] singular Values (n={}, nThr={})".format(Space, Name, len(sig), nOrbExpected))
    print("    [{}]".format(', '.join('{:.4f}'.format(k) for k in sig)))
    print("    Sigma[{}]                         {:.4f} (should be 1)".format(nOrbExpected-1, sig[nOrbExpected-1]))
    if nOrbExpected < len(sig):
        print("    Sigma[{}]                         {:.4f} (should be 0)".format(nOrbExpected, sig[nOrbExpected]))
        if sig[nOrbExpected-1] < 0.8 or sig[nOrbExpected] > 0.5:
            raise Exception("{} orbital construction okay?".format(Space))

    V = Vt.T
    COut = np.dot(COrb, V[:,:nOrbExpected])
    return SemiCanonicalize(COut, Fock, S1, Name)



# Use these to add 'stuff' to the proto-active space.
def MakePiSystemOrbitals(TargetName, iTargetAtomsForPlane_,
                         iTargetAtomsForBasis_,Elements,Coords, CIb, Shells,
                         S1, S12, S2, Fock, COcc, CVir):
    print("\n *** TARGET: {}\n".format(TargetName))
    # convert from 1-based atom indices to 0-based indices.
    iTargetAtomsForPlane = np.array(iTargetAtomsForPlane_) - 1
    if iTargetAtomsForBasis_ is None:
        iTargetAtomsForBasis = iTargetAtomsForPlane
    else:
        iTargetAtomsForBasis = np.array(iTargetAtomsForBasis_) - 1

    vPz = GetPzOrientation(iTargetAtomsForPlane, Coords, Elements)
    CAoMix = MakePzMinaoVectors(iTargetAtomsForBasis, vPz, Shells)

    nPiElec = 0
    for iAt in iTargetAtomsForBasis:
        nPiElec += GetNumPiElec(iAt, Elements,Coords)

    print("    Formal number of elec in pi-system      {} ".format(nPiElec))

    if (nPiElec % 2) != 0:
        raise Exception("Got {} pi electrons. Shouldn't this be an even number?".format(nPiElec))

    # idea: make full pi system for target AOs (via IAOs)
    # then we do know how many occupied and virtual pi
    # orbitals there are supposed to be (by counting the contributed pi
    # electrons from the constituent main group atoms)

    # we then isolate those, canonicalize them, and take
    # the frontier MO subset.

    CTargetIb = np.dot(CIb, CAoMix)
    nTargetIb = CAoMix.shape[1]
    print("    Number of target MINAO basis fn      {} ".format(nTargetIb))

    print("    size of CIb1          {} ".format(CIb.shape[0]))
    print("    size of CAomix1       {} ".format(CAoMix.shape[0]))
    print("    size of CIb2          {} ".format(CIb.shape[1]))
    print("    size of CAomix2       {} ".format(CAoMix.shape[1]))


    nOccOrbExpected=nPiElec//2
    nVirtOrbExpected=nTargetIb - nPiElec//2
    CPiOcc = MakeOverlappingOrbSubspace("Pi", "Occ", COcc, nOccOrbExpected,   CTargetIb, S1, Fock)
    CPiVir = MakeOverlappingOrbSubspace("Pi", "Vir", CVir, nVirtOrbExpected,   CTargetIb, S1, Fock)
    return CPiOcc, CPiVir,nOccOrbExpected,nVirtOrbExpected



# Copyright 2023 The GPU4PySCF Authors. All Rights Reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Hessian of SMD solvent model (for experiment and education)
copied from GPU4PySCF with modification for CPU
'''

import numpy as np
from pyscf.solvent.grad import smd as smd_grad
from pyscf.solvent.smd_experiment import (
    sigma_water, sigma_n, sigma_alpha, sigma_beta, r_zz, swtich_function)

def hess_swtich_function(R, r, dr):
    if R < r + dr:
        dist = (R[0]**2 + R[1]**2 + R[2]**2)**.5
        hess = [
            [R[1]**2+R[2]**2, -R[0]*R[1], -R[0]*R[2]],
            [-R[0]*R[1], R[0]**2+R[2]**2, -R[1]*R[2]],
            [-R[0]*R[2], -R[1]*R[2], R[0]**2+R[1]**2]]
        return np.asarray(hess)/dist
    else:
        return np.zeros([3,3])

def atomic_surface_tension(symbols, coords, n, alpha, beta, water=True):
    '''
    TODO: debug later
    - list of atomic symbols
    - atomic coordinates in Anstrong
    - solvent descriptors: n, alpha, beta
    '''

    def get_bond_tension(bond):
        if water:
            return sigma_water.get(bond, 0.0)
        t = 0.0
        t += sigma_n.get(bond, 0.0) * n
        t += sigma_alpha.get(bond, 0.0) * alpha
        t += sigma_beta.get(bond, 0.0) * beta
        return t

    def get_atom_tension(sym_i):
        if water:
            return sigma_water.get(sym_i, 0.0)
        t = 0.0
        t += sigma_n.get(sym_i, 0.0) * n
        t += sigma_alpha.get(sym_i, 0.0) * alpha
        t += sigma_beta.get(sym_i, 0.0) * beta
        return t
    natm = coords.shape[0]
    ri_rj = coords[:,None,:] - coords[None,:,:]
    rij = np.sum(ri_rj**2, axis=2)**0.5
    np.fill_diagonal(rij, 1)
    #drij = ri_rj / np.expand_dims(rij, axis=-1)
    tensions = []
    for i, sym_i in enumerate(symbols):
        if sym_i not in ['H', 'C', 'N', 'O', 'F', 'Si', 'S', 'Cl', 'Br']:
            tensions.append(np.zeros([natm,3]))
            continue

        tension = np.zeros([natm,3])
        if sym_i in ['F', 'Si', 'S', 'Cl', 'Br']:
            tensions.append(tension)
            continue

        if sym_i == 'H':
            dt_HC = np.zeros([natm,natm,3,3])
            dt_HO = np.zeros([natm,natm,3,3])
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C':
                    r, dr = r_zz.get(('H','C'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_HC[i,i] += dt_drij
                    dt_HC[i,j] -= dt_drij
                    dt_HC[j,i] -= dt_drij
                    dt_HC[j,j] += dt_drij
                if sym_j == 'O':
                    r, dr = r_zz.get(('H','O'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_HO[i,i] += dt_drij
                    dt_HO[i,j] -= dt_drij
                    dt_HO[j,i] -= dt_drij
                    dt_HO[j,j] += dt_drij
            sig_HC = get_bond_tension(('H','C'))
            sig_HO = get_bond_tension(('H','O'))
            tension += sig_HC * dt_HC + sig_HO * dt_HO
            tensions.append(tension)

        if sym_i == 'C':
            dt_CN = np.zeros([natm,3])
            d2t_CC = np.zeros([natm,natm,3,3])
            d2t_CN = np.zeros([natm,natm,3,3])
            t_CN = 0.0
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C' and i != j:
                    r, dr = r_zz.get(('C', 'C'), (0.0, 0.0))
                    d2t_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    d2t_CC[i,i] += d2t_drij
                    d2t_CC[i,j] -= d2t_drij
                    d2t_CC[j,i] -= d2t_drij
                    d2t_CC[j,j] += d2t_drij
                if sym_j == 'N':
                    r, dr = r_zz.get(('C', 'N'), (0.0, 0.0))
                    t_CN += swtich_function(rij[i,j], r, dr)
                    dt_drij = smd_grad.grad_switch_function(coords[i]-coords[j], r, dr)
                    dt_CN[i] += dt_drij
                    dt_CN[j] -= dt_drij
                    d2t_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    d2t_CN[i,i] += d2t_drij
                    d2t_CN[i,j] -= d2t_drij
                    d2t_CN[j,i] -= d2t_drij
                    d2t_CN[j,j] += d2t_drij
            sig_CC = get_bond_tension(('C','C'))
            sig_CN = get_bond_tension(('C','N'))
            tension += sig_CC * d2t_CC + sig_CN * (2 * t_CN * d2t_CN + 2 * np.einsum('i,j->ij', dt_drij[i], dt_drij[j]))
            tensions.append(tension)

        if sym_i == 'N':
            t_NC = 0.0
            dt_NC = np.zeros([natm,natm,3,3])
            dt_NC3 = np.zeros([natm,natm,3,3])
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C':
                    r, dr = r_zz.get(('N','C'), (0.0, 0.0))
                    tk = 0.0
                    dtk = np.zeros([natm,natm,3,3])
                    for k, sym_k in enumerate(symbols):
                        if k != i and k != j:
                            rjk, drjk = r_zz.get(('C', sym_k), (0.0, 0.0))
                            tk += swtich_function(rij[j,k], rjk, drjk)
                            dtk_rjk = hess_swtich_function(coords[j]-coords[k], rjk, drjk)
                            dtk[j,j] += dtk_rjk
                            dtk[j,k] -= dtk_rjk
                            dtk[k,j] -= dtk_rjk
                            dtk[k,k] += dtk_rjk
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr) * tk**2
                    dt_NC[i,i] += dt_drij
                    dt_NC[i,j] -= dt_drij
                    dt_NC[j,i] -= dt_drij
                    dt_NC[j,j] += dt_drij
                    t = swtich_function(coords[i]-coords[j], r, dr)
                    dt_NC += t * (2 * tk * dtk)
                    t_NC += t * tk**2

                    r, dr = r_zz.get(('N','C3'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_NC3[i,i] += dt_drij
                    dt_NC3[i,j] -= dt_drij
                    dt_NC3[j,i] -= dt_drij
                    dt_NC3[j,j] += dt_drij
            sig_NC = get_bond_tension(('N','C'))
            sig_NC3= get_bond_tension(('N','C3'))
            tension += sig_NC * (1.3 * t_NC**0.3 * dt_NC) + sig_NC3 * dt_NC3
            tensions.append(tension)

        if sym_i == 'O':
            dt_OC = np.zeros([natm,natm,3,3])
            dt_ON = np.zeros([natm,natm,3,3])
            dt_OO = np.zeros([natm,natm,3,3])
            dt_OP = np.zeros([natm,natm,3,3])
            for j, sym_j in enumerate(symbols):
                if sym_j == 'C':
                    r, dr = r_zz.get(('O','C'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_OC[i,i] += dt_drij
                    dt_OC[i,j] -= dt_drij
                    dt_OC[j,i] -= dt_drij
                    dt_OC[j,j] += dt_drij
                if sym_j == 'N':
                    r, dr = r_zz.get(('O','N'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_ON[i,i] += dt_drij
                    dt_ON[i,j] -= dt_drij
                    dt_ON[j,i] -= dt_drij
                    dt_ON[j,j] += dt_drij
                if sym_j == 'O' and j != i:
                    r, dr = r_zz.get(('O','O'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_OO[i,i] += dt_drij
                    dt_OO[i,j] -= dt_drij
                    dt_OO[j,i] -= dt_drij
                    dt_OO[j,j] += dt_drij
                if sym_j == 'P':
                    r, dr = r_zz.get(('O','P'), (0.0, 0.0))
                    dt_drij = hess_swtich_function(coords[i]-coords[j], r, dr)
                    dt_OP[i,i] += dt_drij
                    dt_OP[i,j] -= dt_drij
                    dt_OP[j,i] -= dt_drij
                    dt_OP[j,j] += dt_drij
            sig_OC = get_bond_tension(('O','C'))
            sig_ON = get_bond_tension(('O','N'))
            sig_OO = get_bond_tension(('O','O'))
            sig_OP = get_bond_tension(('O','P'))
            tension += sig_OC * dt_OC + sig_ON * dt_ON + sig_OO * dt_OO + sig_OP * dt_OP
            tensions.append(tension)
    return np.asarray(tensions)



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#         Susi Lehtola <susi.lehtola@gmail.com>

import numpy
from pyscf import gto
from pyscf.lib import logger
from pyscf.lib import param
from pyscf.data import elements
from pyscf.scf import atom_hf, ADIIS
from pyscf.dft import rks

def get_atm_nrks(mol, atomic_configuration=elements.NRSRHFS_CONFIGURATION, xc='slater', grid=(100, 434)):
    elements = {a[0] for a in mol._atom}
    logger.info(mol, 'Spherically averaged atomic KS for %s', elements)

    atm_template = mol.copy(deep=False)
    atm_template.charge = 0
    atm_template.symmetry = False  # TODO: enable SO3 symmetry here
    atm_template.atom = atm_template._atom = []
    atm_template.cart = False  # AtomSphericAverageRKS does not support cartesian basis

    atm_scf_result = {}
    for ia, a in enumerate(mol._atom):
        element = a[0]
        if element in atm_scf_result:
            continue

        atm = atm_template
        atm._atom = [a]
        atm._atm = mol._atm[ia:ia+1]
        atm._bas = mol._bas[mol._bas[:,0] == ia].copy()
        atm._ecpbas = mol._ecpbas[mol._ecpbas[:,0] == ia].copy()
        # Point to the only atom
        atm._bas[:,0] = 0
        atm._ecpbas[:,0] = 0
        if element in mol._pseudo:
            atm._pseudo = {element: mol._pseudo.get(element)}
            raise NotImplementedError
        atm.spin = atm.nelectron % 2

        nao = atm.nao
        # nao == 0 for the case that no basis was assigned to an atom
        if nao == 0 or atm.nelectron == 0:  # GHOST
            mo_occ = mo_energy = numpy.zeros(nao)
            mo_coeff = numpy.zeros((nao,nao))
            atm_scf_result[element] = (0, mo_energy, mo_coeff, mo_occ)
        else:
            atm_ks = AtomSphericAverageRKS(atm)
            atm_ks.atomic_configuration = atomic_configuration
            atm_ks.xc = xc
            atm_ks.grids.atom_grid = grid
            atm_ks.verbose = mol.verbose
            my_diis_obj = ADIIS()
            my_diis_obj.space = 12
            atm_ks.diis = my_diis_obj
            atm_ks.run()
            atm_scf_result[element] = (atm_ks.e_tot, atm_ks.mo_energy,
                                       atm_ks.mo_coeff, atm_ks.mo_occ)
    return atm_scf_result


class AtomSphAverageRKS(rks.RKS, atom_hf.AtomSphericAverageRHF):
    def __init__(self, mol, *args, **kwargs):
        atom_hf.AtomSphericAverageRHF.__init__(self, mol)
        rks.RKS.__init__(self, mol, *args, **kwargs)

        # SAP guess is perfect for atoms
        self.init_guess = 'vsap'

    eig = atom_hf.AtomSphericAverageRHF.eig

    get_occ = atom_hf.AtomSphericAverageRHF.get_occ

    get_grad = atom_hf.AtomSphericAverageRHF.get_grad

AtomSphericAverageRKS = AtomSphAverageRKS



#!/usr/bin/env python
# Copyright 2020-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import hf as mol_hf
from pyscf.pbc.lib import kpts as libkpts
from pyscf.pbc.scf import khf_ksymm, khf, kuhf
from pyscf.pbc.lib.kpts import KPoints

@lib.with_doc(kuhf.get_occ.__doc__)
def get_occ(mf, mo_energy_kpts=None, mo_coeff_kpts=None):
    if mo_energy_kpts is None:
        mo_energy_kpts = mf.mo_energy
    kpts = mf.kpts
    assert isinstance(kpts, KPoints)

    nocc_a, nocc_b = mf.nelec
    mo_energy_kpts = kpts.transform_mo_energy(mo_energy_kpts)
    mo_energy = np.sort(np.hstack(mo_energy_kpts[0]))
    fermi_a = mo_energy[nocc_a-1]
    mo_occ_kpts = [[], []]
    for mo_e in mo_energy_kpts[0]:
        mo_occ_kpts[0].append((mo_e <= fermi_a).astype(np.double))
    if nocc_a < len(mo_energy):
        logger.info(mf, 'alpha HOMO = %.12g  LUMO = %.12g', fermi_a, mo_energy[nocc_a])
    else:
        logger.info(mf, 'alpha HOMO = %.12g  (no LUMO because of small basis) ', fermi_a)

    if nocc_b > 0:
        mo_energy = np.sort(np.hstack(mo_energy_kpts[1]))
        fermi_b = mo_energy[nocc_b-1]
        for mo_e in mo_energy_kpts[1]:
            mo_occ_kpts[1].append((mo_e <= fermi_b).astype(np.double))
        if nocc_b < len(mo_energy):
            logger.info(mf, 'beta HOMO = %.12g  LUMO = %.12g', fermi_b, mo_energy[nocc_b])
        else:
            logger.info(mf, 'beta HOMO = %.12g  (no LUMO because of small basis) ', fermi_b)
    else:
        for mo_e in mo_energy_kpts[1]:
            mo_occ_kpts[1].append(np.zeros_like(mo_e))

    if mf.verbose >= logger.DEBUG:
        np.set_printoptions(threshold=len(mo_energy))
        logger.debug(mf, '     k-point                  alpha mo_energy')
        for k,kpt in enumerate(mf.cell.get_scaled_kpts(kpts, kpts_in_ibz=False)):
            if (np.count_nonzero(mo_occ_kpts[0][k]) > 0 and
                np.count_nonzero(mo_occ_kpts[0][k] == 0) > 0):
                logger.debug(mf, '  %2d (%6.3f %6.3f %6.3f)   %s %s',
                             k, kpt[0], kpt[1], kpt[2],
                             np.sort(mo_energy_kpts[0][k][mo_occ_kpts[0][k]> 0]),
                             np.sort(mo_energy_kpts[0][k][mo_occ_kpts[0][k]==0]))
            else:
                logger.debug(mf, '  %2d (%6.3f %6.3f %6.3f)   %s',
                             k, kpt[0], kpt[1], kpt[2], mo_energy_kpts[0][k])
        logger.debug(mf, '     k-point                  beta  mo_energy')
        for k,kpt in enumerate(mf.cell.get_scaled_kpts(kpts, kpts_in_ibz=False)):
            if (np.count_nonzero(mo_occ_kpts[1][k]) > 0 and
                np.count_nonzero(mo_occ_kpts[1][k] == 0) > 0):
                logger.debug(mf, '  %2d (%6.3f %6.3f %6.3f)   %s %s',
                             k, kpt[0], kpt[1], kpt[2],
                             np.sort(mo_energy_kpts[1][k][mo_occ_kpts[1][k]> 0]),
                             np.sort(mo_energy_kpts[1][k][mo_occ_kpts[1][k]==0]))
            else:
                logger.debug(mf, '  %2d (%6.3f %6.3f %6.3f)   %s',
                             k, kpt[0], kpt[1], kpt[2], mo_energy_kpts[1][k])
        np.set_printoptions(threshold=1000)

    if isinstance(kpts, KPoints):
        mo_occ_kpts[0] = kpts.check_mo_occ_symmetry(mo_occ_kpts[0], tol=1e-4)
        mo_occ_kpts[1] = kpts.check_mo_occ_symmetry(mo_occ_kpts[1], tol=1e-4)
    return mo_occ_kpts

@lib.with_doc(kuhf.energy_elec.__doc__)
def energy_elec(mf, dm_kpts=None, h1e_kpts=None, vhf_kpts=None):
    if dm_kpts is None: dm_kpts = mf.make_rdm1()
    if h1e_kpts is None: h1e_kpts = mf.get_hcore()
    if vhf_kpts is None: vhf_kpts = mf.get_veff(mf.cell, dm_kpts)
    wtk = mf.kpts.weights_ibz

    e1 = np.einsum('k,kij,kji', wtk, dm_kpts[0], h1e_kpts)
    e1+= np.einsum('k,kij,kji', wtk, dm_kpts[1], h1e_kpts)
    e_coul = np.einsum('k,kij,kji', wtk, dm_kpts[0], vhf_kpts[0]) * 0.5
    e_coul+= np.einsum('k,kij,kji', wtk, dm_kpts[1], vhf_kpts[1]) * 0.5
    mf.scf_summary['e1'] = e1.real
    mf.scf_summary['e2'] = e_coul.real
    logger.debug(mf, 'E1 = %s  E_coul = %s', e1, e_coul)
    if kuhf.CHECK_COULOMB_IMAG and abs(e_coul.imag) > mf.cell.precision*10:
        logger.warn(mf, "Coulomb energy has imaginary part %s. "
                    "Coulomb integrals (e-e, e-N) may not converge !",
                    e_coul.imag)
    return (e1+e_coul).real, e_coul.real

get_rho = khf_ksymm.get_rho


class KsymAdaptedKUHF(khf_ksymm.KsymAdaptedKSCF, kuhf.KUHF):
    """
    KUHF with k-point symmetry
    """

    get_occ = get_occ
    energy_elec = energy_elec
    get_rho = get_rho

    to_ks = kuhf.KUHF.to_ks
    convert_from_ = kuhf.KUHF.convert_from_

    def __init__(self, cell, kpts=libkpts.KPoints(),
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'),
                 use_ao_symmetry=True):
        khf_ksymm.ksymm_scf_common_init(self, cell, kpts, use_ao_symmetry)
        kuhf.KUHF.__init__(self, cell, kpts, exxdiv)

    @property
    def nelec(self):
        if self._nelec is not None:
            return self._nelec
        else:
            cell = self.cell
            nkpts = self.kpts.nkpts
            ne = cell.tot_electrons(nkpts)
            nalpha = (ne + cell.spin) // 2
            nbeta = nalpha - cell.spin
            if nalpha + nbeta != ne:
                raise RuntimeError('Electron number %d and spin %d are not consistent\n'
                                   'Note cell.spin = 2S = Nalpha - Nbeta, not 2S+1' %
                                   (ne, cell.spin))
            return nalpha, nbeta

    @nelec.setter
    def nelec(self, x):
        self._nelec = x

    def dump_flags(self, verbose=None):
        khf_ksymm.KsymAdaptedKSCF.dump_flags(self, verbose)
        logger.info(self, 'number of electrons per unit cell  '
                    'alpha = %d beta = %d', *self.nelec)
        return self

    def get_init_guess(self, cell=None, key='minao', s1e=None):
        if s1e is None:
            s1e = self.get_ovlp(cell)
        dm_kpts = mol_hf.SCF.get_init_guess(self, cell, key)
        assert dm_kpts.shape[0]==2
        if dm_kpts.ndim != 4:
            nkpts = self.kpts.nkpts_ibz
            # dm[spin,nao,nao] at gamma point -> dm_kpts[spin,nkpts,nao,nao]
            dm_kpts = np.repeat(dm_kpts[:,None,:,:], nkpts, axis=1)
        elif dm_kpts.shape[1] != self.kpts.nkpts_ibz:
            dm_kpts = dm_kpts[:,self.kpts.ibz2bz]

        ne = lib.einsum('k,xkij,kji->x', self.kpts.weights_ibz, dm_kpts, s1e).real
        nkpts = self.kpts.nkpts
        ne *= nkpts
        nelec = np.asarray(self.nelec)
        if np.any(abs(ne - nelec) > 0.01*nkpts):
            logger.debug(self, 'Big error detected in the electron number '
                         'of initial guess density matrix (Ne/cell = %g)!\n'
                         '  This can cause huge error in Fock matrix and '
                         'lead to instability in SCF for low-dimensional '
                         'systems.\n  DM is normalized wrt the number '
                         'of electrons %s', ne.mean()/nkpts, nelec/nkpts)
            dm_kpts *= (nelec / ne).reshape(2,-1,1,1)
        return dm_kpts

    def eig(self, h_kpts, s_kpts):
        e_a, c_a = khf_ksymm.KsymAdaptedKSCF.eig(self, h_kpts[0], s_kpts)
        e_b, c_b = khf_ksymm.KsymAdaptedKSCF.eig(self, h_kpts[1], s_kpts)
        return (e_a,e_b), (c_a,c_b)

    def get_orbsym(self, mo_coeff=None, s=None):
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        if s is None:
            s = self.get_ovlp()

        orbsym_a = khf_ksymm.KsymAdaptedKSCF.get_orbsym(self, mo_coeff[0], s)
        orbsym_b = khf_ksymm.KsymAdaptedKSCF.get_orbsym(self, mo_coeff[1], s)
        return (orbsym_a, orbsym_b)

    orbsym = property(get_orbsym)

    def _finalize(self):
        from pyscf.scf import chkfile as mol_chkfile
        kuhf.KUHF._finalize(self)
        if not self.use_ao_symmetry:
            return

        orbsym = self.get_orbsym()
        for s in range(2):
            for k, mo_e in enumerate(self.mo_energy[s]):
                idx = np.argsort(mo_e.round(9), kind='stable')
                self.mo_energy[s][k] = self.mo_energy[s][k][idx]
                self.mo_occ[s][k] = self.mo_occ[s][k][idx]
                self.mo_coeff[s][k] = lib.tag_array(self.mo_coeff[s][k][:,idx],
                                                    orbsym=orbsym[s][k][idx])
        self.dump_chk({'e_tot': self.e_tot, 'mo_energy': self.mo_energy,
                       'mo_coeff': self.mo_coeff, 'mo_occ': self.mo_occ})
        return self

KUHF = KsymAdaptedKUHF
